import requests
from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = ("Set layer index scheme in Solr")

    args = 'path [path...]'

    def handle(self, *args, **options):

        """
        reset core:
        rm -Rf SOLR_HOME/server/solr/hypermap_test
        solr6 create_core -c hypermap_test
        """

        # read more here: https://cwiki.apache.org/confluence/display/solr/Schema+API
        schema_url = '%s/schema' % settings.SEARCH_URL
        schema_url = "http://localhost:8983/solr/hypermap_test/schema"

        # defining fields: https://cwiki.apache.org/confluence/display/solr/Defining+Fields
        # stored and indexed are default true.
        fields = [
            {"name": "abstract", "type": "string"},
            {"name": "area", "type": "tdouble"},
            {"name": "availability", "type": "string"},
            {"name": "bbox", "type": "location_rpt"},
            {"name": "domain_name", "type": "string"},
            {"name": "id", "type": "tlong", "required": True},
            {"name": "is_public", "type": "boolean"},
            {"name": "last_status", "type": "boolean"},
            {"name": "layer_date", "type": "tdate", "docValues": True},
            {"name": "layer_datetype", "type": "string"},
            {"name": "layer_id", "type": "tlong"},
            {"name": "layer_originator", "type": "string"},
            {"name": "location", "type": "string"},
            {"name": "max_x", "type": "tdouble"},
            {"name": "max_y", "type": "tdouble"},
            {"name": "min_x", "type": "tdouble"},
            {"name": "min_y", "type": "tdouble"},
            {"name": "name", "type": "string"},
            {"name": "recent_reliability", "type": "tdouble"},
            {"name": "reliability", "type": "tdouble"},
            {"name": "service_id", "type": "tlong"},
            {"name": "service_type", "type": "string"},
            {"name": "srs", "type": "string", "multiValued": True},
            {"name": "tile_url", "type": "string"},
            {"name": "title", "type": "string"},
            {"name": "type", "type": "string"},
            {"name": "url", "type": "string"},
        ]

        headers = {
            "Content-type": "application/json"
        }

        for field in fields:
            data = {
                "add-field": field
            }
            print data
            res = requests.post(schema_url, json=data, headers=headers)
            print res.text
            print '*' * 20



import json
import requests
import time
from django.conf import settings
from django.core.urlresolvers import reverse
from django.test import TestCase

from hypermap.aggregator.models import Catalog
from hypermap.search_api import utils


class SearchApiTestCase(TestCase):

    def index(self, data):
        headers = {"content-type": "application/json"}
        params = {"commitWithin": 1500}
        solr_json = json.dumps(data)
        requests.post(self.url_solr_update, data=solr_json, params=params, headers=headers)
        # solr seems to have a delay indexing the layers.
        # before to proceed with the tests wait for 2 secs.
        # otherwise it will return zero docs in the next test.
        time.sleep(2)

    def setUp(self):
        self.url_solr_clear = '%s/update?commit=true' % settings.SEARCH_URL
        self.url_solr_update = '%s/update/json/docs' % settings.SEARCH_URL
        self.url_solr_search = '%s/select' % settings.SEARCH_URL
        self.api_url = "http://localhost:8000{0}".format(reverse("search_api",
                                                                 args=["hypermap"]))
        self.default_params = {
            "search_engine": "solr",
            "search_engine_endpoint": self.url_solr_search,
            "q_time": "[* TO *]",
            "q_geo": "[-90,-180 TO 90,180]",
            "d_docs_limit": 0,
            "d_docs_page": 1,
            "d_docs_sort": "score"
        }

        Catalog.objects.get_or_create(
            name="hypermap", slug="hypermap"
        )

        # delete solr documents
        headers = {"content-type": "text/xml"}
        requests.post(self.url_solr_clear,
                      data="<delete><query>*:*</query></delete>",
                      headers=headers)

        # add test solr documents
        self.solr_records = [
            {
                "abstract": "Downtown-New-Orleans-Map ",
                "layer_originator": "warp.worldmap.harvard.edu",
                "availability": "Online",
                "min_x": -40.0,
                "max_x": -20.0,
                "max_y": -20.0,
                "min_y": -40.0,
                "bbox": "ENVELOPE(-40.0,-20.0,-20.0,-40.0)",
                "title": "1",
                "domain_name": "warp.worldmap.harvard.edu",
                "id": 1,
                "location": "{\"layerInfoPage\": \"/layer/998/\"}",
                "service_type": "Hypermap:WARPER",
                "layer_datetype": "From Metadata",
                "layer_id": 1,
                "is_public": True,
                "layer_date": "2000-03-01T00:00:00Z",
                "name": "1490",
                "url": "http://warp.worldmap.harvard.edu/maps/wms/1490?",
                "type": "Layer",
                "area": 1,
                "tile_url": "/layers/998/map/wmts/nypl_map/default_grid/{z}/{y}/{x}.png",
                "srs": ["EPSG:3857",
                        "EPSG:900913",
                        "EPSG:4326"],
                "service_id": 1
            },
            {
                "abstract": "Downtown-New-Orleans-Map ",
                "layer_originator": "warp.worldmap.harvard.edu",
                "availability": "Online",
                "min_x": -40,
                "max_x": -20,
                "max_y": 40,
                "min_y": 20,
                "bbox": "ENVELOPE(-40,-20,40,20)",
                "title": "20",
                "domain_name": "warp.worldmap.harvard.edu",
                "id": 2,
                "location": "{\"layerInfoPage\": \"/layer/998/\"}",
                "service_type": "Hypermap:WARPER",
                "layer_datetype": "From Metadata",
                "layer_id": 2,

                "is_public": True,
                "layer_date": "2001-03-01T00:00:00Z",
                "name": "1490",
                "url": "http://warp.worldmap.harvard.edu/maps/wms/1490?",
                "type": "Layer",
                "area": 2,
                "tile_url": "/layers/998/map/wmts/nypl_map/default_grid/{z}/{y}/{x}.png",
                "srs": ["EPSG:3857",
                        "EPSG:900913",
                        "EPSG:4326"],
                "service_id": 2
            },
            {
                "abstract": "Downtown-New-Orleans-Map ",
                "layer_originator": "warp.worldmap.harvard.edu",
                "availability": "Online",
                "min_x": 20,
                "max_x": 40,
                "max_y": 40,
                "min_y": 20,
                "bbox": "ENVELOPE(20,40,40,20)",
                "title": "3",
                "domain_name": "warp.worldmap.harvard.edu",
                "id": 3,
                "location": "{\"layerInfoPage\": \"/layer/998/\"}",
                "service_type": "Hypermap:WARPER",
                "layer_datetype": "From Metadata",
                "layer_id": 30,
                "is_public": True,
                "layer_date": "2002-03-01T00:00:00Z",
                "name": "1490",
                "url": "http://warp.worldmap.harvard.edu/maps/wms/1490?",
                "type": "Layer",
                "area": 3,
                "tile_url": "/layers/998/map/wmts/nypl_map/default_grid/{z}/{y}/{x}.png",
                "srs": ["EPSG:3857",
                        "EPSG:900913",
                        "EPSG:4326"],
                "service_id": 3
            },
            {
                "abstract": "Downtown-New-Orleans-Map ",
                "layer_originator": "warp.worldmap.harvard.edu",
                "availability": "Online",
                "min_x": 20,
                "max_x": 40,
                "max_y": -20,
                "min_y": -40,
                "bbox": "ENVELOPE(20,40,-20,-40)",
                "title": "Layer with a BCE date",
                "domain_name": "warp.worldmap.harvard.edu",
                "id": 4,
                "location": "{\"layerInfoPage\": \"/layer/998/\"}",
                "service_type": "Hypermap:WARPER",
                "layer_datetype": "From Metadata",
                "layer_id": 4,
                "is_public": True,
                "layer_date": "-5000000-01-01T00:00:00Z",
                "name": "1490",
                "url": "http://warp.worldmap.harvard.edu/maps/wms/1490?",
                "type": "Layer",
                "area": 4,
                "tile_url": "/layers/998/map/wmts/nypl_map/default_grid/{z}/{y}/{x}.png",
                "srs": ["EPSG:3857",
                        "EPSG:900913",
                        "EPSG:4326"],
                "service_id": 4
            }

        ]

        self.index(self.solr_records)

    def test_catalogs(self):
        url = settings.SITE_URL + reverse("catalog-list")
        res = requests.get(url)
        self.assertEqual(res.status_code, 200)
        catalogs = res.json()
        self.assertEqual(len(catalogs), 1)

    def test_all_match_docs(self):
        params = self.default_params
        print "searching on [{}]".format(self.api_url)
        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)

        results = results.json()
        self.assertEqual(results["a.matchDocs"], len(self.solr_records))

    def test_q_text(self):
        params = self.default_params
        params["q_text"] = "title:1"
        params["d_docs_limit"] = 100

        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)

        results = results.json()
        self.assertEqual(results["a.matchDocs"], 1)

        for doc in results.get("d.docs", []):
            self.assertEqual(doc["title"], "1")

    def test_q_time(self):
        params = self.default_params

        # test validations
        params["q_time"] = "[2000-01-01 - 2001-01-01T00:00:00]"
        results = requests.get(self.api_url, params=params)
        # requires [X TO Y]
        self.assertEqual(400, results.status_code)

        # test asterisks
        # all times
        params["q_time"] = "[* TO *]"
        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)
        results = results.json()
        # all records
        self.assertEqual(results["a.matchDocs"], len(self.solr_records))

        # test range
        # entire year 2000
        params["q_time"] = "[2000-01-01 TO 2001-01-01T00:00:00]"
        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)
        results = results.json()
        # 1 in year 2000
        self.assertEqual(results["a.matchDocs"], 1)

        # test complete min and max when q time is asterisks
        params["q_time"] = "[* TO *]"
        params["a_time_limit"] = 1
        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)
        results = results.json()
        self.assertEqual(results["a.matchDocs"], len(self.solr_records))
        # * to first date
        self.assertEqual(results["a.time"]["start"], "-5000000-01-01T00:00:00Z")
        # TODO: returning "2100-01-01T00:00:00Z" because BCE dates cant calculate time duration.
        # self.assertEqual(results["a.time"]["end"], "2002-03-01T00:00:00Z") # * to last date

        # test facets
        params["q_time"] = "[2000 TO 2022]"
        params["a_time_limit"] = 1
        params["a_time_gap"] = "P1Y"
        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)
        results = results.json()
        self.assertEqual(results["a.matchDocs"], 3)
        # 2000 to complete datetime format
        self.assertEqual(results["a.time"]["start"], "2000-01-01T00:00:00Z")
        # 2022 to complete datetime format
        self.assertEqual(results["a.time"]["end"], "2022-01-01T00:00:00Z")
        # the facet counters are all facets excluding < 2000
        self.assertEqual(len(results["a.time"]["counts"]), 3)

    def test_q_geo(self):
        params = self.default_params

        # top right square
        params["q_geo"] = "[0,0 TO 30,30]"
        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)
        results = results.json()
        self.assertEqual(results["a.matchDocs"], 1)

        # bottom left square
        params["q_geo"] = "[-30,-30 TO 0,0]"
        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)
        results = results.json()
        self.assertEqual(results["a.matchDocs"], 1)

        # big square
        params["q_geo"] = "[-30,-30 TO 30,30]"
        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)
        results = results.json()
        self.assertEqual(results["a.matchDocs"], 4)

        # center where no layers
        params["q_geo"] = "[-5,-5 TO 5,5]"
        results = requests.get(self.api_url, params=params)
        self.assertEqual(results.status_code, 200)
        results = results.json()
        self.assertEqual(results["a.matchDocs"], 0)

        # bad format
        params["q_geo"] = "[-5,-5 5,5]"
        results = requests.get(self.api_url, params=params)
        # validate the format
        self.assertEqual(results.status_code, 400)

    def test_utilities(self):
        # test_parse_datetime_range
        start, end = utils.parse_datetime_range("[2013-03-01 TO 2014-05-02T23:00:00]")
        self.assertTrue(start.get("is_common_era"))
        self.assertEqual(start.get("parsed_datetime").year, 2013)
        self.assertEqual(start.get("parsed_datetime").month, 3)
        self.assertEqual(start.get("parsed_datetime").day, 1)
        self.assertTrue(end.get("is_common_era"))
        self.assertEqual(end.get("parsed_datetime").year, 2014)
        self.assertEqual(end.get("parsed_datetime").month, 5)
        self.assertEqual(end.get("parsed_datetime").day, 2)
        self.assertEqual(end.get("parsed_datetime").hour, 23)
        self.assertEqual(end.get("parsed_datetime").minute, 0)
        self.assertEqual(end.get("parsed_datetime").second, 0)

        start, end = utils.parse_datetime_range("[-500000000 TO 2014-05-02T23:00:00]")
        self.assertFalse(start.get("is_common_era"))
        self.assertEqual(start.get("parsed_datetime"), "-500000000-01-01T00:00:00Z")

        start, end = utils.parse_datetime_range("[* TO *]")
        self.assertTrue(start.get("is_common_era"))
        self.assertEqual(start.get("parsed_datetime"), None)
        self.assertEqual(end.get("parsed_datetime"), None)

        # test_parse_ISO8601
        quantity, units = utils.parse_ISO8601("P3D")
        self.assertEqual(quantity, 3)
        self.assertEqual(units[0], "DAYS")

        # test_gap_to_sorl
        value = utils.gap_to_sorl("P3D")
        self.assertEqual(value, "+3DAYS")

        # test_parse_geo_box
        value = utils.parse_geo_box("[-90,-180 TO 90,180]")
        self.assertEqual(value.bounds[0], -90)
        self.assertEqual(value.bounds[1], -180)
        self.assertEqual(value.bounds[2], 90)
        self.assertEqual(value.bounds[3], 180)

        # test_request_time_facet
        d = utils.request_time_facet("x", "[2000 TO 2014-01-02T11:12:13]", None, 1000)
        self.assertEqual(type(d), dict)
        self.assertEqual(d['f.x.facet.range.start'], '2000-01-01T00:00:00Z')
        self.assertEqual(d['f.x.facet.range.end'], '2014-01-02T11:12:13Z')
        self.assertEqual(d['f.x.facet.range.gap'], '+6DAYS')
        self.assertEqual(d['facet.range'], 'x')

        d = utils.request_time_facet("y", "[-5000000 TO 2016]", "P1D", 1)
        self.assertEqual(d['f.y.facet.range.start'], '-5000000-01-01T00:00:00Z')
        self.assertEqual(d['f.y.facet.range.end'], '2016-01-01T00:00:00Z')
        self.assertEqual(d['f.y.facet.range.gap'], '+1DAYS')
        self.assertEqual(d['facet.range'], 'y')



from django.conf import settings

from models import Service
from utils import create_services_from_endpoint


def populate_initial_services():
    """
    Populate a fresh installed Hypermap instances with basic services.
    """
    services_list = (
        (
            'Harvard WorldMap',
            'Harvard WorldMap open source web geospatial platform',
            'Hypermap:WorldMap',
            'http://worldmap.harvard.edu'
        ),
        (
            'NYPL MapWarper',
            'The New York Public Library (NYPL) MapWarper web site',
            'Hypermap:WARPER',
            'http://maps.nypl.org/warper/maps'
        ),
        (
            'Map Warper',
            'The MapWarper web site developed, hosted and maintained by Tim Waters',
            'Hypermap:WARPER',
            'http://mapwarper.net/maps'
        ),
        (
            'WorldMap Warp',
            'The MapWarper instance part of the Harvard WorldMap project',
            'Hypermap:WARPER',
            'http://warp.worldmap.harvard.edu/maps'
        ),
        (
            'WFP GeoNode',
            'World Food Programme GeoNode',
            'OGC:WMS',
            'http://geonode.wfp.org/geoserver/ows?'
        ),
        (
            'NASA EARTHDATA',
            'NASA EARTHDATA, powered by EOSDIS',
            'OGC:WMTS',
            'http://map1.vis.earthdata.nasa.gov/wmts-geo/1.0.0/WMTSCapabilities.xml'
        ),
    )

    esri_endpoint = 'https://gis.ngdc.noaa.gov/arcgis/rest/services'
    print '*** Importing esri endpoint: %s' % esri_endpoint
    create_services_from_endpoint(esri_endpoint)

    for service in services_list:
        print '*** Importing %s' % service[0]
        service = Service(
            title=service[0],
            abstract=service[1],
            type=service[2],
            url=service[3]
        )
        service.save()

settings.SKIP_CELERY_TASK = True
populate_initial_services()



from __future__ import absolute_import

from django.conf import settings

from celery import shared_task

REGISTRY_LIMIT_LAYERS = getattr(settings, 'REGISTRY_LIMIT_LAYERS', -1)
REGISTRY_SEARCH_URL = getattr(settings, 'REGISTRY_SEARCH_URL', None)

if REGISTRY_SEARCH_URL is None:
    SEARCH_ENABLED = False
    SEARCH_TYPE = None
    SEARCH_URL = None
else:
    SEARCH_ENABLED = True
    SEARCH_TYPE = REGISTRY_SEARCH_URL.split('+')[0]
    SEARCH_URL = REGISTRY_SEARCH_URL.split('+')[1]


if REGISTRY_LIMIT_LAYERS > 0:
    DEBUG_SERVICES = True
    DEBUG_LAYERS_NUMBER = REGISTRY_LIMIT_LAYERS
else:
    DEBUG_SERVICES = False
    DEBUG_LAYERS_NUMBER = -1


@shared_task(bind=True)
def check_all_services(self):
    from hypermap.aggregator.models import Service
    service_to_processes = Service.objects.filter(active=True)
    total = service_to_processes.count()
    count = 0
    for service in service_to_processes:
        # update state
        if not self.request.called_directly:
            self.update_state(
                state='PROGRESS',
                meta={'current': count, 'total': total}
            )
        check_service.delay(service)
        count = count + 1


@shared_task(bind=True)
def check_service(self, service):
    # total is determined (and updated) exactly after service.update_layers
    total = 100

    def status_update(count):
        if not self.request.called_directly:
            self.update_state(
                state='PROGRESS',
                meta={'current': count, 'total': total}
            )

    status_update(0)
    service.update_layers()
    service.index_layers()
    # we count 1 for update_layers and 1 for service check for simplicity
    layer_to_process = service.layer_set.all()

    if DEBUG_SERVICES:
        layer_to_process = layer_to_process[0:DEBUG_LAYERS_NUMBER]

    total = layer_to_process.count() + 2
    status_update(1)
    service.check_available()
    status_update(2)
    count = 3

    if not settings.REGISTRY_SKIP_CELERY:
        for layer in layer_to_process:
            # update state
            status_update(count)
            check_layer.delay(layer)
            count += 1
    else:
        for layer in layer_to_process:
            status_update(count)
            check_layer(layer)
            count += 1


@shared_task(bind=True, time_limit=10)
def check_layer(self, layer):
    print 'Checking layer %s' % layer.name
    success, message = layer.check_available()
    # every time a layer is checked it should be indexed
    if success and SEARCH_ENABLED:
        if not settings.REGISTRY_SKIP_CELERY:
            index_layer.delay(layer)
        else:
            index_layer(layer)
    if not success:
        from hypermap.aggregator.models import TaskError
        task_error = TaskError(
            task_name=self.name,
            args=layer.id,
            message=message
        )
        task_error.save()


@shared_task(name="clear_index")
def clear_index():
    if SEARCH_TYPE == 'solr':
        print 'Clearing the solr core and indexes'
        from hypermap.aggregator.solr import SolrHypermap
        solrobject = SolrHypermap()
        solrobject.clear_solr()
    elif SEARCH_TYPE == 'elasticsearch':
        print 'Clearing the ES indexes'
        from hypermap.aggregator.elasticsearch_client import ESHypermap
        esobject = ESHypermap()
        esobject.clear_es()


@shared_task(bind=True)
def remove_service_checks(self, service):

    service.check_set.all().delete()

    def status_update(count, total):
        if not self.request.called_directly:
            self.update_state(
                state='PROGRESS',
                meta={'current': count, 'total': total}
            )

    layer_to_process = service.layer_set.all()
    count = 0
    total = layer_to_process.count()
    for layer in layer_to_process:
        # update state
        status_update(count, total)
        layer.check_set.all().delete()
        count = count + 1


@shared_task(bind=True)
def index_service(self, service):

    layer_to_process = service.layer_set.all()
    total = layer_to_process.count()

    def status_update(count):
        if not self.request.called_directly:
            self.update_state(
                state='PROGRESS',
                meta={'current': count, 'total': total}
            )

    count = 0
    for layer in layer_to_process:
        # update state
        status_update(count)
        if not settings.REGISTRY_SKIP_CELERY:
            index_layer.delay(layer)
        else:
            index_layer(layer)
        count = count + 1


@shared_task(bind=True)
def index_layer(self, layer):
    # TODO: Make this function more DRY
    # by abstracting the common bits.
    if SEARCH_TYPE == 'solr':
        from hypermap.aggregator.solr import SolrHypermap
        print 'Syncing layer %s to solr' % layer.name
        try:
            solrobject = SolrHypermap()
            success, message = solrobject.layer_to_solr(layer)
            if not success:
                from hypermap.aggregator.models import TaskError
                task_error = TaskError(
                    task_name=self.name,
                    args=layer.id,
                    message=message
                )
                task_error.save()
        except:
            print 'There was an exception here!'
            self.retry(layer)
    elif SEARCH_TYPE == 'elasticsearch':
        from hypermap.aggregator.elasticsearch_client import ESHypermap
        print 'Syncing layer %s to es' % layer.name
        esobject = ESHypermap()
        success, message = esobject.layer_to_es(layer)
        if not success:
            from hypermap.aggregator.models import TaskError
            task_error = TaskError(
                task_name=self.name,
                args=layer.id,
                message=message
            )
            task_error.save()


@shared_task(bind=True)
def index_all_layers(self):
    from hypermap.aggregator.models import Layer

    layer_to_processes = Layer.objects.all()
    total = layer_to_processes.count()
    count = 0
    for layer in Layer.objects.all():
        # update state
        if not self.request.called_directly:
            self.update_state(
                state='PROGRESS',
                meta={'current': count, 'total': total}
            )
        if not settings.REGISTRY_SKIP_CELERY:
            index_layer.delay(layer)
        else:
            index_layer(layer)
        count = count + 1


@shared_task(bind=True)
def update_endpoint(self, endpoint, greedy_opt=False):
    from hypermap.aggregator.utils import create_services_from_endpoint
    from hypermap.aggregator.models import Endpoint

    print 'Processing endpoint with id %s: %s' % (endpoint.id, endpoint.url)

    # Override the greedy_opt var with the value from the endpoint list
    # if it's available.
    if endpoint.endpoint_list:
        greedy_opt = endpoint.endpoint_list.greedy

    imported, message = create_services_from_endpoint(endpoint.url, greedy_opt=greedy_opt, catalog=endpoint.catalog)

    # this update will not execute the endpoint_post_save signal.
    Endpoint.objects.filter(id=endpoint.id).update(
        imported=imported, message=message, processed=True
    )


@shared_task(bind=True)
def update_endpoints(self, endpoint_list):
    # for now we process the enpoint even if they were already processed
    endpoint_to_process = endpoint_list.endpoint_set.filter(processed=False)
    total = endpoint_to_process.count()
    count = 0
    if not settings.REGISTRY_SKIP_CELERY:
        for endpoint in endpoint_to_process:
            update_endpoint.delay(endpoint)
        # update state
        if not self.request.called_directly:
            self.update_state(
                state='PROGRESS',
                meta={'current': count, 'total': total}
            )
    else:
        for endpoint in endpoint_to_process:
            update_endpoint(endpoint)

    return True



from django.conf import settings
from urlparse import urlparse


def resource_urls(request):
    """Global values to pass to templates"""
    REGISTRY_SEARCH_URL = getattr(settings, "REGISTRY_SEARCH_URL", "elasticsearch+http://localhost:9200")

    SEARCH_TYPE = REGISTRY_SEARCH_URL.split('+')[0]
    SEARCH_URL = REGISTRY_SEARCH_URL.split('+')[1]

    url_parsed = urlparse(SEARCH_URL)
    defaults = dict(
        SITE_URL=settings.SITE_URL.rstrip('/'),
        SEARCH_TYPE=SEARCH_TYPE,
        SEARCH_URL=SEARCH_URL,
        SEARCH_IP='%s://%s:%s' % (url_parsed.scheme, url_parsed.hostname, url_parsed.port)
    )
    return defaults



import sys
import logging
import math
import json

from urlparse import urlparse
from django.conf import settings
from django.utils.html import strip_tags

from elasticsearch import Elasticsearch
from shapely.geometry import box

from hypermap.aggregator.utils import mercator_to_llbbox

from hypermap.aggregator.solr import get_date


REGISTRY_MAPPING_PRECISION = getattr(settings, "REGISTRY_MAPPING_PRECISION", "500m")
REGISTRY_SEARCH_URL = getattr(settings, "REGISTRY_SEARCH_URL", "elasticsearch+http://localhost:9200")

SEARCH_TYPE = REGISTRY_SEARCH_URL.split('+')[0]
SEARCH_URL = REGISTRY_SEARCH_URL.split('+')[1]


class ESHypermap(object):

    es_url = SEARCH_URL
    es = Elasticsearch(hosts=[es_url])
    index_name = 'hypermap'
    logger = logging.getLogger("hypermap")

    def __init__(self):
        # TODO: this create_indices() should not happen here:
        # ES creates the indexes automaticaly.
        super(ESHypermap, self).__init__()

    @staticmethod
    def good_coords(coords):
        """ passed a string array """
        if (len(coords) != 4):
            return False
        for coord in coords[0:3]:
            try:
                num = float(coord)
                if (math.isnan(num)):
                    return False
                if (math.isinf(num)):
                    return False
            except ValueError:
                return False
        return True

    @staticmethod
    def get_domain(url):
        urlParts = urlparse(url)
        hostname = urlParts.hostname
        if hostname == "localhost":
            return "Harvard"  # assumption
        return hostname

    @staticmethod
    def get_bbox(layer):
        candidate_bbox = layer.bbox_x0, layer.bbox_y0, layer.bbox_x1, layer.bbox_y1
        if None not in candidate_bbox:
            return [float(coord) for coord in candidate_bbox]

        wkt = layer.wkt_geometry
        # If a coordinate is None and 'POLYGON'
        if 'POLYGON' in wkt:
            from shapely.wkt import loads
            return loads(wkt).bounds

        return (-180.0, -90.0, 180.0, 90.0)

    @staticmethod
    def layer_to_es(layer):
        category = None
        username = None
        ESHypermap.logger.info("Elasticsearch: record to save: [%s] %s" % (layer.catalog.slug, layer.id))

        try:
            bbox = ESHypermap.get_bbox(layer)
            for proj in layer.service.srs.values():
                if proj['code'] in ('102113', '102100'):
                    bbox = mercator_to_llbbox(bbox)
            if (ESHypermap.good_coords(bbox)) is False:
                print 'Elasticsearch: There are not valid coordinates for this layer ', layer.title
                ESHypermap.logger.error('Elasticsearch: There are not valid coordinates for layer id: %s' % layer.id)
                return False
            if (ESHypermap.good_coords(bbox)):
                minX = bbox[0]
                minY = bbox[1]
                maxX = bbox[2]
                maxY = bbox[3]
                if (minY > maxY):
                    minY, maxY = maxY, minY
                if (minX > maxX):
                    minX, maxX = maxX, minX
                halfWidth = (maxX - minX) / 2.0
                halfHeight = (maxY - minY) / 2.0
                area = (halfWidth * 2) * (halfHeight * 2)
                if (minX < -180):
                    minX = -180
                if (maxX > 180):
                    maxX = 180
                if (minY < -90):
                    minY = -90
                if (maxY > 90):
                    maxY = 90
                wkt = "ENVELOPE({:f},{:f},{:f},{:f})".format(minX, maxX, maxY, minY)
                rectangle = box(minX, minY, maxX, maxY)
                domain = ESHypermap.get_domain(layer.service.url)
                if hasattr(layer, 'layerwm'):
                    category = layer.layerwm.category
                    username = layer.layerwm.username
                abstract = layer.abstract
                if abstract:
                    abstract = strip_tags(layer.abstract)
                else:
                    abstract = ''
                if layer.service.type == "WM":
                    originator = username
                else:
                    originator = domain
                # we need to remove the exising index in case there is already one
                # ESHypermap.es.delete('hypermap', 'layer', layer.id)
                # now we add the index
                es_record = {
                    "id": str(layer.id),
                    "type": 'Layer',
                    "layer_id": str(layer.id),
                    "name": layer.name,
                    "title": layer.title,
                    "layer_originator": originator,
                    "service_id": str(layer.service.id),
                    "service_type": layer.service.type,
                    "layer_category": category,
                    "layer_username": username,
                    "url": layer.url,
                    "reliability": layer.reliability,
                    "recent_reliability": layer.recent_reliability,
                    "last_status": layer.last_status,
                    "is_public": layer.is_public,
                    "availability": "Online",
                    "location": '{"layer_info": "' + layer.get_absolute_url + '"}',
                    "abstract": abstract,
                    "domain_name": layer.service.get_domain,
                    # "SrsProjectionCode": layer.srs.values_list('code', flat=True),
                    "min_y": minY,
                    "min_x": minX,
                    "max_x": maxX,
                    "max_y": maxY,
                    "area": area,
                    "bbox": wkt,
                    "centroid_x": rectangle.centroid.x,
                    "centroid_y": rectangle.centroid.y,
                    "srs": [srs.encode('utf-8') for srs in layer.service.srs.values_list('code', flat=True)],
                    "layer_geoshape": {
                        "type": "envelope",
                        "coordinates": [
                            [minX, maxY], [maxX, minY]
                        ]
                    },
                }

                es_date, type = get_date(layer)
                if es_date is not None:
                    es_record['layer_date'] = es_date
                    es_record['layer_datetype'] = type

                if layer.get_tile_url():
                    es_record['tile_url'] = layer.get_tile_url()

                ESHypermap.logger.info(es_record)
                # TODO: cache index creation.
                ESHypermap.create_indices(layer.catalog.slug)
                ESHypermap.es.index(layer.catalog.slug, 'layer', json.dumps(es_record), id=layer.id,
                                    request_timeout=20)
                ESHypermap.logger.info("Elasticsearch: record saved for layer with id: %s" % layer.id)
                return True, None
        except Exception:
            ESHypermap.logger.error(sys.exc_info())
            ESHypermap.logger.error("Elasticsearch: Error saving record for layer with id: %s - %s"
                                    % (layer.id, sys.exc_info()[1]))
            return False, sys.exc_info()[1]

    @staticmethod
    def clear_es():
        """Clear all indexes in the es core"""
        # TODO: should receive a catalog slug.
        ESHypermap.es.indices.delete(ESHypermap.index_name, ignore=[400, 404])
        print 'Elasticsearch: Index cleared'

    @staticmethod
    def create_indices(catalog_slug):
        """Create ES core indices """
        # TODO: enable auto_create_index in the ES nodes to make this implicit.
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-index_.html#index-creation
        # http://support.searchly.com/customer/en/portal/questions/
        # 16312889-is-automatic-index-creation-disabled-?new=16312889
        mapping = {
            "mappings": {
                "layer": {
                    "properties": {
                        "layer_geoshape": {
                            "type": "geo_shape",
                            "tree": "quadtree",
                            "precision": REGISTRY_MAPPING_PRECISION
                        }
                    }
                }
            }
        }
        ESHypermap.es.indices.create(catalog_slug, ignore=[400, 404], body=mapping)



from django.conf.urls import url, include
from django.views.generic import TemplateView
from rest_framework import routers
from hypermap.search_api import views


router = routers.DefaultRouter()
router.register(r'catalogs', views.CatalogViewSet)


urlpatterns = [
    url(r'^api/', include(router.urls)),
    url(r'^api/docs/$', TemplateView.as_view(template_name='search_api/swagger/index.html')),
    url(r'^(?P<catalog_slug>[-\w]+)/api/$', views.Search.as_view(), name="search_api"),
]



from hypermap.aggregator.models import Catalog
from . import utils
from rest_framework import serializers


class SearchSerializer(serializers.Serializer):

    # TODO: remove this after catalogs integration:
    search_engine = serializers.ChoiceField(
        required=False,
        help_text="Where will be running the search.",
        choices=["solr", "elasticsearch"]
    )
    search_engine_endpoint = serializers.URLField(
        required=False,
        help_text="Endpoint URL",
    )
    # TODO:END.

    q_time = serializers.CharField(
        required=False,
        help_text="Constrains docs by time range. Either side can be '*' to signify open-ended. "
                  "Otherwise it must be in either format as given in the example. UTC time zone is implied. Example: "
                  "[2013-03-01 TO 2013-04-01T00:00:00]"
    )
    q_geo = serializers.CharField(
        required=False,
        help_text="A rectangular geospatial filter in decimal degrees going from the lower-left to the upper-right. "
                  "The coordinates are in lat,lon format. "
                  "Example: [-90,-180 TO 90,180]"
    )
    q_text = serializers.CharField(
        required=False,
        help_text="Constrains docs by keyword search query."
    )
    q_user = serializers.CharField(
        required=False,
        help_text="Constrains docs by matching exactly a certain user."
    )
    d_docs_limit = serializers.IntegerField(
        required=False,
        help_text="How many documents to return.",
        default=0
    )
    d_docs_page = serializers.IntegerField(
        required=False,
        help_text="When documents to return are more than d_docs_limit they can be paginated by this value.",
        default=1
    )
    d_docs_sort = serializers.ChoiceField(
        required=False,
        help_text="How to order the documents before returning the top X. 'score' is keyword search relevancy. "
                  "'time' is time descending. 'distance' is the distance between the doc and the middle of q.geo.",
        default="score",
        choices=["score", "time", "distance"]
    )
    a_time_limit = serializers.IntegerField(
        required=False,
        help_text="Non-0 triggers time/date range faceting. This value is the maximum number of time ranges to "
                  "return when a.time.gap is unspecified. This is a soft maximum; less will usually be returned. "
                  "A suggested value is 100. Note that a.time.gap effectively ignores this value. "
                  "See Solr docs for more details on the query/response format.",
        default=0
    )
    a_time_gap = serializers.CharField(
        required=False,
        help_text="The consecutive time interval/gap for each time range. Ignores a.time.limit.The format is based on "
                  "a subset of the ISO-8601 duration format."
    )
    a_time_filter = serializers.CharField(
        required=False,
        help_text=("From what time range to divide by a.time.gap into intervals."
                   "Defaults to q.time and otherwise 90 days.")
    )

    a_hm_limit = serializers.IntegerField(
        required=False,
        help_text=("Non-0 triggers heatmap/grid faceting. "
                   "This number is a soft maximum on thenumber of cells it should have. "
                   "There may be as few as 1/4th this number in return. "
                   "Note that a.hm.gridLevel can effectively ignore this value. "
                   "The response heatmap contains a counts grid that can be null or contain null rows when "
                   "all its values would be 0. See Solr docs for more details on the response format."),
        default=0
    )
    a_hm_gridlevel = serializers.IntegerField(
        required=False,
        help_text="To explicitly specify the grid level, e.g. to let a user ask for greater or courser resolution "
                  "than the most recent request. Ignores a.hm.limit."
    )
    a_hm_filter = serializers.CharField(
        required=False,
        help_text="To explicitly specify the grid level, e.g. to let a user ask for greater or courser resolution "
                  "than the most recent request. Ignores a.hm.limit."
    )

    a_text_limit = serializers.IntegerField(
        required=False,
        help_text="Returns the most frequently occurring words. WARNING: There is usually a significant performance "
                  "hit in this due to the extremely high cardinality.",
        default=0
    )
    a_user_limit = serializers.IntegerField(
        required=False,
        help_text="Returns the most frequently occurring users.",
        default=0
    )
    original_response = serializers.IntegerField(
        required=False,
        help_text="Returns te original search engine response.",
        default=0
    )

    def validate_q_time(self, value):
        """
        Would be for example: [2013-03-01 TO 2013-04-01T00:00:00] and/or [* TO *]
        Returns a valid sorl value. [2013-03-01T00:00:00Z TO 2013-04-01T00:00:00Z] and/or [* TO *]
        """
        if value:
            try:
                range = utils.parse_datetime_range_to_solr(value)
                return range
            except Exception as e:
                raise serializers.ValidationError(e.message)

        return value

    def validate_q_geo(self, value):
        """
        Would be for example: [-90,-180 TO 90,180]
        """
        if value:
            try:
                rectangle = utils.parse_geo_box(value)
                return "[{0},{1} TO {2},{3}]".format(
                    rectangle.bounds[0],
                    rectangle.bounds[1],
                    rectangle.bounds[2],
                    rectangle.bounds[3],
                )
            except Exception as e:
                raise serializers.ValidationError(e.message)

        return value

    def validate_a_time_filter(self, value):
        """
        Would be for example: [2013-03-01 TO 2013-04-01:00:00:00] and/or [* TO *]
        """
        if value:
            try:
                utils.parse_datetime_range(value)
            except Exception as e:
                raise serializers.ValidationError(e.message)

        return value

    def validate_d_docs_page(self, value):
        """
        paginations cant be zero or negative.
        :param value:
        :return:
        """
        if value <= 0:
            raise serializers.ValidationError("d_docs_page cant be zero or negative")
        return value


class CatalogSerializer(serializers.HyperlinkedModelSerializer):
    search_url = serializers.CharField(source="get_search_url",
                                       read_only=True)

    class Meta:
        model = Catalog
        fields = ('id', 'slug', 'name', 'url', 'search_url')



import sys
import pysolr
import requests
import logging
import json
import datetime

from urlparse import urlparse
from django.conf import settings
from django.utils.html import strip_tags
from shapely.geometry import box

from hypermap.aggregator.utils import mercator_to_llbbox


def get_date(layer):
    """
    Returns a date for Solr. A date can be detected or from metadata.
    It can be a range or a simple date in isoformat.
    """
    date = None
    sign = '+'
    date_type = 1
    layer_dates = layer.get_layer_dates()
    # we index the first date!
    if layer_dates:
        sign = layer_dates[0][0]
        date = layer_dates[0][1]
        date_type = layer_dates[0][2]
    if date is None:
        date = layer.created.date()
    # layer date > 2300 is invalid for sure
    # TODO put this logic in date miner
    if date.year > 2300:
        date = None
    if date_type == 0:
        date_type = "Detected"
    if date_type == 1:
        date_type = "From Metadata"
    return get_solr_date(date, (sign == '-')), date_type


def get_solr_date(pydate, is_negative):
    """
    Returns a date in a valid Solr format from a string.
    """
    # check if date is valid and then set it to solr format YYYY-MM-DDThh:mm:ssZ
    try:
        if isinstance(pydate, datetime.datetime):
            solr_date = '%sZ' % pydate.isoformat()[0:19]
            if is_negative:
                print '***** This layer has a negative date'
                solr_date = '-%s' % solr_date
            return solr_date
        else:
            return None
    except Exception:
        return None


class SolrHypermap(object):

    def get_domain(self, url):
        urlParts = urlparse(url)
        hostname = urlParts.hostname
        if hostname == "localhost":
            return "Harvard"  # assumption
        return hostname

    def layer_to_solr(self, layer):
        logger = logging.getLogger("hypermap")
        category = None
        username = None
        try:
            bbox = None
            if not layer.has_valid_bbox():
                message = 'There are not valid coordinates for layer id: %s' % layer.id
                logger.error(message)
            else:
                bbox = [float(layer.bbox_x0), float(layer.bbox_y0), float(layer.bbox_x1), float(layer.bbox_y1)]
                for proj in layer.service.srs.values():
                    if proj['code'] in ('102113', '102100'):
                        bbox = mercator_to_llbbox(bbox)
                minX = bbox[0]
                minY = bbox[1]
                maxX = bbox[2]
                maxY = bbox[3]
                # coords hack needed by solr
                if (minX < -180):
                    minX = -180
                if (maxX > 180):
                    maxX = 180
                if (minY < -90):
                    minY = -90
                if (maxY > 90):
                    maxY = 90
                wkt = "ENVELOPE({:f},{:f},{:f},{:f})".format(minX, maxX, maxY, minY)
                halfWidth = (maxX - minX) / 2.0
                halfHeight = (maxY - minY) / 2.0
                area = (halfWidth * 2) * (halfHeight * 2)
                rectangle = box(minX, minY, maxX, maxY)
            domain = self.get_domain(layer.service.url)
            if hasattr(layer, 'layerwm'):
                category = layer.layerwm.category
                username = layer.layerwm.username
            abstract = layer.abstract
            if abstract:
                abstract = strip_tags(layer.abstract)
            else:
                abstract = ''
            if layer.type == "WM":
                originator = username
            else:
                originator = domain
            # now we add the index
            solr_record = {
                           'id': layer.id,
                           'type': 'Layer',
                           'layer_id': layer.id,
                           'name': layer.name,
                           'title': layer.title,
                           'layer_originator': originator,
                           'service_id': layer.service.id,
                           'service_type': layer.service.type,
                           'layer_category': category,
                           'layer_username': username,
                           'url': layer.url,
                           'reliability': layer.reliability,
                           'recent_reliability': layer.recent_reliability,
                           'last_status': layer.last_status,
                           'is_public': layer.is_public,
                           'availability': 'Online',
                           'location': '{"layerInfoPage": "' + layer.get_absolute_url + '"}',
                           'abstract': abstract,
                           'domain_name': layer.service.get_domain
                           }

            solr_date, date_type = get_date(layer)
            if solr_date is not None:
                solr_record['layer_date'] = solr_date
                solr_record['layer_datetype'] = date_type
            if bbox is not None:
                solr_record['min_x'] = minX
                solr_record['min_y'] = minY
                solr_record['max_x'] = maxX
                solr_record['max_y'] = maxY
                solr_record['area'] = area
                solr_record['bbox'] = wkt
                solr_record['centroid_x'] = rectangle.centroid.x
                solr_record['centroid_y'] = rectangle.centroid.y

                srs_list = [srs.encode('utf-8') for srs in layer.service.srs.values_list('code', flat=True)]
                # solr_record['srs'] = ', '.join(srs_list)
                solr_record['srs'] = srs_list
            if layer.get_tile_url():
                solr_record['tile_url'] = layer.get_tile_url()

            # time to send request to solr
            url_solr_update = '%s/update/json/docs' % settings.SEARCH_URL
            headers = {"content-type": "application/json"}
            params = {"commitWithin": 1500}
            solr_json = json.dumps(solr_record)
            requests.post(url_solr_update, data=solr_json, params=params,  headers=headers)
            logger.info("Solr record saved for layer with id: %s" % layer.id)
            return True, None
        except Exception:
            logger.error("Error saving solr record for layer with id: %s - %s" % (layer.id, sys.exc_info()[1]))
            return False, sys.exc_info()[1]

    def clear_solr(self):
        """Clear all indexes in the solr core"""
        solr_url = settings.SEARCH_URL
        solr = pysolr.Solr(solr_url, timeout=60)
        solr.delete(q='*:*')
        print 'Solr core cleared'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from django.test import TestCase
from hypermap.dynasty.utils import mine_date


class DateMinerTest(TestCase):

    def setUp(self):
        self.text_plain = 'mytext'
        self.text_year = '2003_Pathv5.0daynight_SST'
        self.text_ce = '1950CE title'
        self.text_bce = '11BCE abstract'
        self.text_dynasty = 'Ming regions'
        self.text_unicode = 'Carte des Etats-unis : 2010 provinces méridionales.'
        self.text_less_resource = "Garden in Front of Palace of Fine Arts<div> " \
                                  "<img src='http://www.ndl.go.jp/site_nippon/viennae/data/img/2_0006.jpg'" \
                                  "align='none'/></div>"
        self.text_bce_three = '160 BCE'
        self.text_multiple_dates = '2003_this_has two 2013'
        self.text_multiple_dates_2 = '_1981 2003_this_has three 2013'
        self.text_bc = '19BC abstract'
        self.text_exact_date = 'My birthday is 02/06/1971 in case you cannot remember'
        self.text_exact_dates = 'My birthday is 02/06/1971 in case you cannot remember' \
                                'My wife birthday is 09/30/1974 and my father was born in 1933.'

    def test_mine_date_none(self):
        self.assertIsNone(mine_date(self.text_plain))

    def test_mine_date_text(self):
        self.assertEqual(mine_date(self.text_year)[0], '2003-01-01')

    def test_year_miner_ce(self):
        self.assertEqual(mine_date(self.text_ce)[0], ['1950-01-01'])

    def test_year_miner_bc(self):
        self.assertEqual(mine_date(self.text_bc)[0], ['-0019-01-01'])

    def test_year_miner_bce_units(self):
        self.assertEqual(mine_date(self.text_bce_three)[0], ['-0160-01-01'])

    def test_year_miner_bce(self):
        self.assertEqual(mine_date(self.text_bce)[0], ['-0011-01-01'])

    def test_year_range_miner_range(self):
        self.assertEqual(mine_date(self.text_dynasty)[0], ['1368-01-01', '1644-01-01'])

    def test_unicode_errors(self):
        self.assertEqual(mine_date(self.text_unicode)[0], '2010-01-01')

    def test_mine_date_none_resource(self):
        self.assertIsNone(mine_date(self.text_less_resource))

    def test_mine_date_multipe(self):
        self.assertEqual(mine_date(self.text_multiple_dates), ['2003-01-01', '2013-01-01'])
        self.assertEqual(mine_date(self.text_multiple_dates_2), ['1981-01-01', '2003-01-01', '2013-01-01'])

    def test_exact_dates(self):
        self.assertEqual(mine_date(self.text_exact_dates), ['1971-01-01', '1974-01-01', '1933-01-01'])



import datetime
import os
import re
import json
import urllib2
import requests
from urlparse import urlparse
from dateutil.parser import parse

from django.conf import settings
from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from django.db.models import Avg, Min, Max
from django.db.models import signals
from django.core.files.uploadedfile import SimpleUploadedFile
from django.core.urlresolvers import reverse
from django_extensions.db.fields import AutoSlugField

from taggit.managers import TaggableManager
from lxml import etree
from shapely.wkt import loads
from owslib.namespaces import Namespaces
from owslib.util import nspath_eval
from owslib.csw import CatalogueServiceWeb
from owslib.tms import TileMapService
from owslib.wms import WebMapService
from owslib.wmts import WebMapTileService
from arcrest import MapService as ArcMapService, ImageService as ArcImageService

from enums import CSW_RESOURCE_TYPES, SERVICE_TYPES, DATE_TYPES
from tasks import update_endpoint, update_endpoints, check_service, check_layer, index_layer
from utils import get_esri_extent, get_esri_service_name, format_float, flip_coordinates

from hypermap.dynasty.utils import get_mined_dates

REGISTRY_LIMIT_LAYERS = getattr(settings, 'REGISTRY_LIMIT_LAYERS', -1)

if REGISTRY_LIMIT_LAYERS > 0:
    DEBUG_SERVICES = True
    DEBUG_LAYER_NUMBER = REGISTRY_LIMIT_LAYERS
else:
    DEBUG_SERVICES = False
    DEBUG_LAYER_NUMBER = -1


def get_parsed_date(sdate):
    try:
        pydate = parse(sdate, yearfirst=True)
        # parser has a problem for dates from 1 to 100
        if sdate[:2] == '00' or sdate[:3] == '000':
            effective_year = int(sdate[:4])
            effective_pydate = datetime.datetime(effective_year, pydate.month, pydate.day)
            pydate = effective_pydate
        return pydate
    except:
        return None


def add_metadata_dates_to_layer(dates, layer):
    default = datetime.datetime(2016, 1, 1)
    for date in dates:
        if date:
            date = '%s' % date
            if date != '':
                if date.startswith('-'):
                    layerdate, created = LayerDate.objects.get_or_create(layer=layer, date=date, type=1)
                else:
                    try:
                        dt = parse(date, default=default)
                        if dt:
                            iso = dt.isoformat()
                            tokens = iso.strip().split("T")
                            fdate = tokens[0]
                            print 'Adding date %s to layer %s' % (fdate, layer.id)
                            layerdate, created = LayerDate.objects.get_or_create(layer=layer, date=fdate, type=1)
                        else:
                            print 'Skipping date "%s" as is invalid.' % date
                    except:
                        print 'Skipping date "%s" as is invalid.' % date


def add_mined_dates(layer):
    text_to_mine = ''
    if layer.title:
        text_to_mine = text_to_mine + layer.title
    if layer.abstract:
        text_to_mine = text_to_mine + ' ' + layer.abstract
    mined_dates = get_mined_dates(text_to_mine)
    for date in mined_dates:
        layer.layerdate_set.get_or_create(date=date, type=0)


class SpatialReferenceSystem(models.Model):
    """
    SpatialReferenceSystem represents a spatial reference system.
    """
    code = models.CharField(max_length=255, null=True, blank=True)

    def __unicode__(self):
        return self.code


class Check(models.Model):
    """
    Check represents the measurement of resource (service/layer) state.
    """
    content_object = generic.GenericForeignKey('content_type', 'object_id')
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    checked_datetime = models.DateTimeField(auto_now=True)
    success = models.BooleanField(default=False)
    response_time = models.FloatField()
    message = models.TextField(default='OK')

    def __unicode__(self):
        return 'Check %s' % self.id


class Resource(models.Model):
    """
    Resource represents basic information for a resource (service/layer).
    """
    title = models.CharField(max_length=255, null=True, blank=True)
    abstract = models.TextField(null=True, blank=True)
    keywords = TaggableManager(blank=True)
    created = models.DateTimeField(auto_now_add=True)
    last_updated = models.DateTimeField(auto_now=True)
    active = models.BooleanField(default=True)
    url = models.URLField(max_length=255)
    is_public = models.BooleanField(default=True)
    type = models.CharField(max_length=32, choices=SERVICE_TYPES)

    check_set = generic.GenericRelation(Check, object_id_field='object_id')

    temporal_extent_start = models.CharField(max_length=255, null=True, blank=True)
    temporal_extent_end = models.CharField(max_length=255, null=True, blank=True)

    # CSW fields
    csw_last_updated = models.CharField(max_length=32,
                                        default=datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ'),
                                        null=True, blank=True)
    csw_type = models.CharField(max_length=32, default='dataset', null=False)
    csw_typename = models.CharField(max_length=32, default='csw:Record', null=False)

    csw_schema = models.CharField(max_length=64,
                                  default='http://www.opengis.net/cat/csw/2.0.2',
                                  null=False)

    anytext = models.TextField(null=True, blank=True)
    wkt_geometry = models.TextField(null=False,
                                    default='POLYGON((-180 -90,-180 90,180 90,180 -90,-180 -90))')

    # metadata XML specific fields
    xml = models.TextField(null=True,
                           default='<csw:Record xmlns:csw="http://www.opengis.net/cat/2.0.2"/>',
                           blank=True)

    def __unicode__(self):
        return str(self.id)

    class Meta:
        abstract = True

    @property
    def id_string(self):
        return str(self.id)

    @property
    def keywords_csv(self):
        keywords_qs = self.keywords.all()
        if keywords_qs:
            return ','.join([kw.name for kw in keywords_qs])
        else:
            return ''

    @property
    def last_updated_iso8601(self):
        if self.last_updated:
            iso8601 = self.last_updated.isoformat()
            if not self.last_updated.utcoffset:
                return '%sZ' % iso8601
            return iso8601

    @property
    def csw_resourcetype(self):
        return CSW_RESOURCE_TYPES[self.type]

    @property
    def first_check(self):
        return self.check_set.order_by('checked_datetime')[0].checked_datetime

    @property
    def last_check(self):
        if self.check_set.all().count() > 0:
            return self.check_set.order_by('-checked_datetime')[0].checked_datetime
        else:
            return None

    @property
    def average_response_time(self):
        # TODO: exclude failed checks with response time = 0.0
        return self.check_set.aggregate(Avg('response_time')).values()[0]

    @property
    def min_response_time(self):
        # TODO: exclude failed checks with response time = 0.0
        return self.check_set.aggregate(Min('response_time')).values()[0]

    @property
    def max_response_time(self):
        # TODO: exclude failed checks with response time = 0.0
        return self.check_set.aggregate(Max('response_time')).values()[0]

    @property
    def last_response_time(self):
        if self.checks_count > 0:
            return self.check_set.order_by('-checked_datetime')[0].response_time
        else:
            return None

    @property
    def last_status(self):
        if self.checks_count > 0:
            return self.check_set.order_by('-checked_datetime')[0].success
        else:
            return None

    @property
    def checks_count(self):
        return self.check_set.all().count()

    @property
    def reliability(self):
        total_checks = self.check_set.count()
        if total_checks:
            success_checks = self.check_set.filter(success=True).count()
            return (success_checks/float(total_checks)) * 100
        else:
            return None

    @property
    def recent_reliability(self):
        total_checks = self.check_set.count()
        recent_checks_number = 2
        if total_checks >= recent_checks_number:
            recent_checks = self.check_set.all().order_by('-checked_datetime')[0:recent_checks_number]
            success_checks = sum(check.success for check in recent_checks)
            return (success_checks/float(recent_checks_number)) * 100
        else:
            return self.reliability

    def get_checks_admin_url(self):
        path = reverse("admin:%s_%s_changelist" % (self._meta.app_label, "check"))
        return path

    def get_checks_admin_reliability_warning_url(self):
        """
        When service Realiability is going down users should go to the
        the check history to find problem causes.
        :return: admin url with check list for this instance
        """
        # TODO: cache this.
        path = self.get_checks_admin_url()
        content_type = ContentType.objects.get_for_model(self)
        params = "?content_type__id__exact={0}&q={1}&success__exact=0".format(
            content_type.id,
            self.id
        )
        url = path + params
        return url


class Service(Resource):
    """
    Service represents a remote geowebservice.
    """

    srs = models.ManyToManyField(SpatialReferenceSystem, blank=True)
    catalog = models.ForeignKey("Catalog", editable=False, default=1)

    @property
    def get_domain(self):
        parsed_uri = urlparse(self.url)
        domain = '{uri.netloc}'.format(uri=parsed_uri)
        return domain

    @property
    def get_absolute_url(self):
        return reverse("service_detail", args=[self.catalog.slug, self.id])

    def update_layers(self):
        """
        Update layers for a service.
        """
        signals.post_save.disconnect(layer_post_save, sender=Layer)
        print 'Updating layers for service id %s' % self.id
        if self.type == 'OGC:WMS':
            update_layers_wms(self)
        elif self.type == 'OGC:WMTS':
            update_layers_wmts(self)
        elif self.type == 'ESRI:ArcGIS:MapServer':
            update_layers_esri_mapserver(self)
        elif self.type == 'ESRI:ArcGIS:ImageServer':
            update_layers_esri_imageserver(self)
        elif self.type == 'Hypermap:WorldMap':
            update_layers_wm(self)
        elif self.type == 'Hypermap:WARPER':
            update_layers_warper(self)
        signals.post_save.connect(layer_post_save, sender=Layer)

    def index_layers(self):
        """
        Index all layers for this service.
        """
        if settings.REGISTRY_SEARCH_URL is not None:
            for layer in self.layer_set.all():
                index_layer(layer)

    def check_available(self):
        """
        Check for availability of a service and provide run metrics.
        """
        success = True
        start_time = datetime.datetime.utcnow()
        message = ''

        print 'Checking service id %s' % self.id

        try:
            title = None
            abstract = None
            keywords = []
            wkt_geometry = None
            srs = '4326'
            if self.type == 'OGC:CSW':
                ows = CatalogueServiceWeb(self.url)
                title = ows.identification.title
                abstract = ows.identification.abstract
                keywords = ows.identification.keywords
            if self.type == 'OGC:WMS':
                ows = WebMapService(self.url)
                title = ows.identification.title
                abstract = ows.identification.abstract
                keywords = ows.identification.keywords
                for c in ows.contents:
                    if ows.contents[c].parent is None:
                        wkt_geometry = bbox2wktpolygon(ows.contents[c].boundingBoxWGS84)
                    break
            if self.type == 'OGC:WMTS':
                ows = WebMapTileService(self.url)
                title = ows.identification.title
                abstract = ows.identification.abstract
                keywords = ows.identification.keywords
            if self.type == 'OSGeo:TMS':
                ows = TileMapService(self.url)
                title = ows.identification.title
                abstract = ows.identification.abstract
                keywords = ows.identification.keywords
            if self.type == 'ESRI:ArcGIS:MapServer':
                esri = ArcMapService(self.url)
                extent, srs = get_esri_extent(esri)
                title = esri.mapName
                if len(title) == 0:
                    title = get_esri_service_name(self.url)
                wkt_geometry = bbox2wktpolygon([
                    extent['xmin'],
                    extent['ymin'],
                    extent['xmax'],
                    extent['ymax']
                ])
            if self.type == 'ESRI:ArcGIS:ImageServer':
                esri = ArcImageService(self.url)
                extent, srs = get_esri_extent(esri)
                title = esri._json_struct['name']
                if len(title) == 0:
                    title = get_esri_service_name(self.url)
                wkt_geometry = bbox2wktpolygon([
                    extent['xmin'],
                    extent['ymin'],
                    extent['xmax'],
                    extent['ymax']
                ])
            if self.type == 'Hypermap:WorldMap':
                urllib2.urlopen(self.url)
                title = 'Harvard WorldMap'
            if self.type == 'Hypermap:WARPER':
                urllib2.urlopen(self.url)
            # update title without raising a signal and recursion
            if title:
                self.title = title
                Service.objects.filter(id=self.id).update(title=title)
            if abstract:
                self.abstract = abstract
                Service.objects.filter(id=self.id).update(abstract=abstract)
            if keywords:
                for kw in keywords:
                    # FIXME: persist keywords to Django model
                    self.keywords.add(kw)
            if wkt_geometry:
                self.wkt_geometry = wkt_geometry
                Service.objects.filter(id=self.id).update(wkt_geometry=wkt_geometry)
            xml = create_metadata_record(
                identifier=self.id_string,
                source=self.url,
                links=[[self.type, self.url]],
                format=self.type,
                type='service',
                title=title,
                abstract=abstract,
                keywords=keywords,
                wkt_geometry=self.wkt_geometry,
                srs=srs
            )
            anytexts = gen_anytext(title, abstract, keywords)
            Service.objects.filter(id=self.id).update(anytext=anytexts, xml=xml, csw_type='service')
        except Exception as err:
            print(err)
            message = str(err)
            success = False

        end_time = datetime.datetime.utcnow()
        delta = end_time - start_time
        response_time = '%s.%s' % (delta.seconds, delta.microseconds)

        check = Check(
            content_object=self,
            success=success,
            response_time=response_time,
            message=message
        )
        check.save()
        print 'Service checked in %s seconds, status is %s' % (response_time, success)


class Catalog(models.Model):
    """
    Represents a collection of layers to be searched.
    """
    name = models.CharField(
        max_length=255,
        help_text="Display name in UI"
    )
    slug = AutoSlugField(
        populate_from='name',
        editable=True,
        help_text="Leave empty to be populated from name"
    )
    url = models.URLField(
        max_length=255,
        help_text=("Only if remote. URL where the API for the search backend is served."
                   "ex: http://localhost:8000/registry/api/search/"),
        null=True, blank=True
    )

    def __unicode__(self):
        return '{0}'.format(
            self.name
        )

    def get_search_url(self):
        """
        resolve the search url no matter if local or remote.
        :return: url or exception
        """

        if self.is_remote:
            return self.url

        return reverse('search_api', args=[self.slug])

    @property
    def is_remote(self):
        """
        Define if data source is local or remote
        :return: True if remote, else False.
        """
        if self.url and len(self.url) > 0:
            return True
        return False

    def get_absolute_url(self):
        return reverse('index', args=[self.slug])


class Layer(Resource):
    """
    Service represents a remote layer.
    """
    name = models.CharField(max_length=255, null=True, blank=True)
    # bbox should be in WGS84
    bbox_x0 = models.DecimalField(max_digits=19, decimal_places=10, blank=True, null=True)
    bbox_x1 = models.DecimalField(max_digits=19, decimal_places=10, blank=True, null=True)
    bbox_y0 = models.DecimalField(max_digits=19, decimal_places=10, blank=True, null=True)
    bbox_y1 = models.DecimalField(max_digits=19, decimal_places=10, blank=True, null=True)
    thumbnail = models.ImageField(upload_to='layers', blank=True, null=True)
    page_url = models.URLField(max_length=255, blank=True, null=True)
    service = models.ForeignKey(Service, blank=True, null=True)
    is_monitored = models.BooleanField(default=True)
    catalog = models.ForeignKey(Catalog, editable=False, default=1)

    def __unicode__(self):
        return '%s' % self.id

    def get_url_endpoint(self):
        """
        Returns the Hypermap endpoint for a layer.
        This endpoint will be the WMTS MapProxy endpoint, only for WM we use the original endpoint.
        """
        endpoint = self.url
        if self.type not in ('Hypermap:WorldMap',):
            endpoint = '%s/registry/%s/layer/%s/map/wmts/1.0.0/WMTSCapabilities.xml' % (
                settings.SITE_URL.rstrip('/'),
                self.catalog.slug,
                self.id
            )
        return endpoint

    def get_tile_url(self):
        """
        Returns the tile url MapProxy endpoint for the layer.
        """
        if self.type not in ('Hypermap:WorldMap',):
            return '/registry/%s/layer/%s/map/wmts/%s/default_grid/{z}/{y}/{x}.png' % (
                   self.catalog.slug, self.id, self.name)
        else:
            return None

    def has_valid_bbox(self):
        if self.bbox_x0 is None or self.bbox_y0 is None or self.bbox_x1 is None or self.bbox_y1 is None:
            return False
        else:
            if self.bbox_x0 > self.bbox_x1 or self.bbox_y0 > self.bbox_y1:
                return False
            else:
                return True

    def get_layer_dates(self):

        def get_date_sign(a_date):
            sign = '+'
            if sdate[0] == '-':
                sign = '-'
            return sign

        # string must be parsed, and check if they have a negative value in front of it
        dates = []
        # get dates from layerwm
        if hasattr(self, 'layerwm'):
            if self.layerwm.temporal_extent_start:
                sdate = self.layerwm.temporal_extent_start
                pydate = get_parsed_date(sdate)
                if pydate:
                    start_date = []
                    start_date.append(get_date_sign(sdate))
                    start_date.append(pydate)
                    start_date.append(1)
                    dates.append(start_date)
            if self.layerwm.temporal_extent_end:
                sdate = self.layerwm.temporal_extent_end
                pydate = get_parsed_date(sdate)
                if pydate:
                    end_date = []
                    end_date.append(get_date_sign(sdate))
                    end_date.append(pydate)
                    end_date.append(1)
                    dates.append(end_date)
        # now we return all the other dates
        for layerdate in self.layerdate_set.all().order_by('date'):
            sdate = layerdate.date
            # for now we skip ranges
            if 'TO' not in sdate:
                pydate = get_parsed_date(sdate)
                if pydate:
                    date = []
                    date.append(get_date_sign(sdate))
                    date.append(pydate)
                    date.append(layerdate.type)
                    dates.append(date)
        print dates
        return dates

    def update_thumbnail(self):
        print 'Generating thumbnail for layer id %s' % self.id
        if not self.has_valid_bbox():
            raise ValueError('Extent for this layer is invalid, cannot generate thumbnail')
            return None
        format_error_message = 'This layer does not expose valid formats (png, jpeg) to generate the thumbnail'
        img = None
        if self.type == 'OGC:WMS':
            ows = WebMapService(self.service.url)
            op_getmap = ows.getOperationByName('GetMap')
            image_format = 'image/png'
            if image_format not in op_getmap.formatOptions:
                if 'image/jpeg' in op_getmap.formatOptions:
                    image_format = 'image/jpeg'
                else:
                    raise NotImplementedError(format_error_message)
            img = ows.getmap(
                layers=[self.name],
                srs='EPSG:4326',
                bbox=(
                    float(self.bbox_x0),
                    float(self.bbox_y0),
                    float(self.bbox_x1),
                    float(self.bbox_y1)
                ),
                size=(50, 50),
                format=image_format,
                transparent=True
            )
            if 'ogc.se_xml' in img.info()['Content-Type']:
                raise ValueError(img.read())
                img = None
        elif self.type == 'OGC:WMTS':

            ows = WebMapTileService(self.service.url)
            ows_layer = ows.contents[self.name]
            image_format = 'image/png'
            if image_format not in ows_layer.formats:
                if 'image/jpeg' in ows_layer.formats:
                    image_format = 'image/jpeg'
                else:
                    raise NotImplementedError(format_error_message)
            img = ows.gettile(
                            layer=self.name,
                            tilematrixset=ows_layer.tilematrixsets[0],
                            tilematrix='0',
                            row='0',
                            column='0',
                            format=image_format
                            )
        elif self.type == 'Hypermap:WorldMap':
            ows = WebMapService(self.url,
                                username=settings.REGISTRY_WORLDMAP_USERNAME,
                                password=settings.REGISTRY_WORLDMAP_PASSWORD)
            op_getmap = ows.getOperationByName('GetMap')
            image_format = 'image/png'
            if image_format not in op_getmap.formatOptions:
                if 'image/jpeg' in op_getmap.formatOptions:
                    image_format = 'image/jpeg'
                else:
                    raise NotImplementedError(format_error_message)
            img = ows.getmap(
                layers=[self.name],
                srs='EPSG:4326',
                bbox=(
                    float(self.bbox_x0),
                    float(self.bbox_y0),
                    float(self.bbox_x1),
                    float(self.bbox_y1)
                ),
                size=(50, 50),
                format=image_format,
                transparent=True
            )
            if 'ogc.se_xml' in img.info()['Content-Type']:
                raise ValueError(img.read())
                img = None
        elif self.type == 'Hypermap:WARPER':
            ows = WebMapService(self.url)
            op_getmap = ows.getOperationByName('GetMap')
            image_format = 'image/png'
            if image_format not in op_getmap.formatOptions:
                if 'image/jpeg' in op_getmap.formatOptions:
                    image_format = 'image/jpeg'
                else:
                    raise NotImplementedError(format_error_message)
            img = ows.getmap(
                layers=[self.name],
                srs='EPSG:4326',
                bbox=(
                    float(self.bbox_x0),
                    float(self.bbox_y0),
                    float(self.bbox_x1),
                    float(self.bbox_y1)
                ),
                size=(50, 50),
                format=image_format,
                transparent=True
            )
            if 'ogc.se_xml' in img.info()['Content-Type']:
                raise ValueError(img.read())
                img = None
        elif self.type == 'ESRI:ArcGIS:MapServer':
            try:
                image = None
                arcserver = ArcMapService(self.service.url)
                bbox = '%s, %s, %s, %s' % (
                    float(self.bbox_x0),
                    float(self.bbox_y0),
                    float(self.bbox_x1),
                    float(self.bbox_y1)
                )

                image = arcserver.ExportMap(
                    bbox=bbox,
                    layers='show:' + self.name,
                    transparent='true',
                    dpi='96',
                    format='jpg'
                )
            except Exception, e:
                print e
            name = re.sub('[^\w\-_\. ]', '_', self.name)
            thumbnail_file_name = '%s%s.jpg' % ('/tmp/', name)
            image.save(thumbnail_file_name)
            img = open(thumbnail_file_name)
            os.remove(thumbnail_file_name)
        elif self.type == 'ESRI:ArcGIS:ImageServer':
            image = None
            try:
                arcserver = ArcImageService(self.service.url)
                bbox = (
                    str(self.bbox_x0) + ',' +
                    str(self.bbox_y0) + ',' +
                    str(self.bbox_x1) + ',' +
                    str(self.bbox_y1)
                )
                image = arcserver.ExportImage(bbox=bbox)
            except Exception, e:
                print e
            name = re.sub('[^\w\-_\. ]', '_', self.name)
            thumbnail_file_name = '%s%s.jpg' % ('/tmp/', name)
            image.save(thumbnail_file_name)
            img = open(thumbnail_file_name)
            os.remove(thumbnail_file_name)

        # update thumb in model
        if img and hasattr(img, 'close'):
            thumbnail_file_name = '%s.jpg' % self.name
            upfile = SimpleUploadedFile(thumbnail_file_name, img.read(), "image/jpeg")
            self.thumbnail.save(thumbnail_file_name, upfile, True)
            print 'Thumbnail updated for layer %s' % self.name

    def check_available(self):
        """
        Check for availability of a layer and provide run metrics.
        """
        success = True
        start_time = datetime.datetime.utcnow()
        message = ''
        print 'Checking layer id %s' % self.id
        try:
            signals.post_save.disconnect(layer_post_save, sender=Layer)
            self.update_thumbnail()
            if settings.REGISTRY_SEARCH_URL is not None:
                if not settings.REGISTRY_SKIP_CELERY:
                    index_layer.delay(self)
                else:
                    index_layer(self)
            signals.post_save.connect(layer_post_save, sender=Layer)

        except ValueError, err:
            # caused by update_thumbnail()
            # self.href is empty in arcserver.ExportMap
            if str(err).startswith("unknown url type:"):
                print 'Thumbnail can not be updated: %s' % str(err)
        except Exception, err:
            message = str(err)
            success = False

        end_time = datetime.datetime.utcnow()

        delta = end_time - start_time
        response_time = '%s.%s' % (delta.seconds, delta.microseconds)

        check = Check(
            content_object=self,
            success=success,
            response_time=response_time,
            message=message
        )
        check.save()
        print 'Layer checked in %s seconds, status is %s' % (response_time, success)
        return success, message

    @property
    def get_absolute_url(self):
        return '{0}{1}'.format(
            settings.SITE_URL,
            reverse("layer_detail", args=[self.catalog.slug, self.id])
        )


class LayerDate(models.Model):
    """
    LayerDate represents list of dates that can be used to depict a layer.
    """
    date = models.CharField(max_length=25)
    type = models.IntegerField(choices=DATE_TYPES)
    layer = models.ForeignKey(Layer)

    def __unicode__(self):
        return self.date


class LayerWM(models.Model):
    """
    LayerWM represents the extended attributes that are found in a WorldMap layer.
    """
    category = models.CharField(max_length=255, null=True, blank=True)
    username = models.CharField(max_length=255, null=True, blank=True)
    temporal_extent_start = models.CharField(max_length=255, null=True, blank=True)
    temporal_extent_end = models.CharField(max_length=255, null=True, blank=True)
    layer = models.OneToOneField(Layer)

    def __unicode__(self):
        return self.layer.name

    class Meta:
        verbose_name = 'WorldMap Layer Attributes'
        verbose_name_plural = 'WorldMap Layers Attributes'


class EndpointList(models.Model):
    """
    EndpointList represents a file containing an EndPoint list.
    """
    upload = models.FileField(upload_to='endpoint_lists')
    greedy = models.BooleanField(default=False)
    catalog = models.ForeignKey(Catalog, editable=False, default=1)

    def __unicode__(self):
        return self.upload.name

    def endpoints_admin_url(self):
        url = '<a href="/admin/aggregator/endpoint/?endpoint_list=%s">Endpoints for this list</a>' % self.id
        return url
    endpoints_admin_url.allow_tags = True


class Endpoint(models.Model):
    """
    Endpoint represents a url containing an end point and its status.
    """
    processed = models.BooleanField(default=False)
    processed_datetime = models.DateTimeField(auto_now=True)
    imported = models.BooleanField(default=False)
    message = models.TextField(blank=True, null=True)
    url = models.URLField(max_length=255)
    endpoint_list = models.ForeignKey(EndpointList, blank=True, null=True)
    catalog = models.ForeignKey(Catalog, editable=False, default=1)

    class Meta:
        unique_together = ("url", "catalog")

    @property
    def id_string(self):
        return str(self.id)


class TaskError(models.Model):
    """
    TaskError represents a task error, until we find a better way to handle this with Celery.
    """
    task_name = models.CharField(max_length=255)
    args = models.CharField(max_length=255)
    error_datetime = models.DateTimeField(auto_now=True)
    message = models.TextField(blank=True, null=True)


def bbox2wktpolygon(bbox):
    """
    Return OGC WKT Polygon of a simple bbox list
    """

    minx = float(bbox[0])
    miny = float(bbox[1])
    maxx = float(bbox[2])
    maxy = float(bbox[3])
    return 'POLYGON((%.2f %.2f, %.2f %.2f, %.2f %.2f, %.2f %.2f, %.2f %.2f))' \
        % (minx, miny, minx, maxy, maxx, maxy, maxx, miny, minx, miny)


def create_metadata_record(**kwargs):
    """
    Create a csw:Record XML document from harvested metadata
    """

    if 'srs' in kwargs:
        srs = kwargs['srs']
    else:
        srs = '4326'

    modified = '%sZ' % datetime.datetime.utcnow().isoformat().split('.')[0]

    nsmap = Namespaces().get_namespaces(['csw', 'dc', 'dct', 'ows'])

    e = etree.Element(nspath_eval('csw:Record', nsmap), nsmap=nsmap)

    etree.SubElement(e, nspath_eval('dc:identifier', nsmap)).text = kwargs['identifier']
    etree.SubElement(e, nspath_eval('dc:title', nsmap)).text = kwargs['title']
    if 'alternative' in kwargs:
        etree.SubElement(e, nspath_eval('dct:alternative', nsmap)).text = kwargs['alternative']
    etree.SubElement(e, nspath_eval('dct:modified', nsmap)).text = modified
    etree.SubElement(e, nspath_eval('dct:abstract', nsmap)).text = kwargs['abstract']
    etree.SubElement(e, nspath_eval('dc:type', nsmap)).text = kwargs['type']
    etree.SubElement(e, nspath_eval('dc:format', nsmap)).text = kwargs['format']
    etree.SubElement(e, nspath_eval('dc:source', nsmap)).text = kwargs['source']

    if 'relation' in kwargs:
        etree.SubElement(e, nspath_eval('dc:relation', nsmap)).text = kwargs['relation']

    if 'keywords' in kwargs:
        if kwargs['keywords'] is not None:
            for keyword in kwargs['keywords']:
                etree.SubElement(e, nspath_eval('dc:subject', nsmap)).text = keyword

    for link in kwargs['links']:
        etree.SubElement(e, nspath_eval('dct:references', nsmap), scheme=link[0]).text = link[1]

    bbox2 = loads(kwargs['wkt_geometry']).bounds
    bbox = etree.SubElement(e, nspath_eval('ows:BoundingBox', nsmap),
                            crs='http://www.opengis.net/def/crs/EPSG/0/%s' % srs,
                            dimensions='2')

    etree.SubElement(bbox, nspath_eval('ows:LowerCorner', nsmap)).text = '%s %s' % (bbox2[1], bbox2[0])
    etree.SubElement(bbox, nspath_eval('ows:UpperCorner', nsmap)).text = '%s %s' % (bbox2[3], bbox2[2])

    return etree.tostring(e, pretty_print=True)


def gen_anytext(*args):
    """
    Convenience function to create bag of words for anytext property
    """

    bag = []

    for term in args:
        if term is not None:
            if isinstance(term, list):
                for term2 in term:
                    if term2 is not None:
                        bag.append(term2)
            else:
                bag.append(term)
    return ' '.join(bag)


# updatelayers for each service type

def update_layers_wms(service):
    """
    Update layers for an OGC:WMS service.
    Sample endpoint: http://demo.geonode.org/geoserver/ows
    """
    try:
        wms = WebMapService(service.url)
        layer_names = list(wms.contents)
        parent = wms.contents[layer_names[0]].parent
        # fallback, some endpoint like this one:
        # https://nsidc.org/cgi-bin/atlas_north?service=WMS&request=GetCapabilities&version=1.1.1
        # does not have a parent to check for srs
        if parent:
            crsOptions = parent.crsOptions
        else:
            crsOptions = wms.contents[layer_names[0]].crsOptions
        # set srs
        for crs_code in crsOptions:
            srs, created = SpatialReferenceSystem.objects.get_or_create(code=crs_code)
            service.srs.add(srs)

        # now update layers
        layer_n = 0
        total = len(layer_names)
        for layer_name in layer_names:
            ows_layer = wms.contents[layer_name]
            print 'Updating layer %s' % ows_layer.name
            # get or create layer
            layer, created = Layer.objects.get_or_create(name=ows_layer.name, service=service, catalog=service.catalog)
            if layer.active:
                links = [['OGC:WMS', service.url]]
                # update fields
                layer.type = 'OGC:WMS'
                layer.title = ows_layer.title
                layer.abstract = ows_layer.abstract
                layer.url = service.url
                layer.page_url = layer.get_absolute_url
                links.append([
                    'WWW:LINK',
                    settings.SITE_URL.rstrip('/') + layer.page_url
                ])
                # bbox
                bbox = list(ows_layer.boundingBoxWGS84 or (-179.0, -89.0, 179.0, 89.0))
                layer.bbox_x0 = bbox[0]
                layer.bbox_y0 = bbox[1]
                layer.bbox_x1 = bbox[2]
                layer.bbox_y1 = bbox[3]
                layer.wkt_geometry = bbox2wktpolygon(bbox)
                # keywords
                for keyword in ows_layer.keywords:
                    layer.keywords.add(keyword)
                # crsOptions
                # TODO we may rather prepopulate with fixutres the SpatialReferenceSystem table
                layer.xml = create_metadata_record(
                    identifier=layer.id_string,
                    source=service.url,
                    links=links,
                    format='OGC:WMS',
                    type=layer.csw_type,
                    relation=service.id_string,
                    title=ows_layer.title,
                    alternative=ows_layer.name,
                    abstract=ows_layer.abstract,
                    keywords=ows_layer.keywords,
                    wkt_geometry=layer.wkt_geometry
                )
                layer.anytext = gen_anytext(layer.title, layer.abstract, ows_layer.keywords)
                layer.save()
                # dates
                add_mined_dates(layer)
            layer_n = layer_n + 1
            # exits if DEBUG_SERVICES
            print "Updating layer n. %s/%s" % (layer_n, total)
            if DEBUG_SERVICES and layer_n == DEBUG_LAYER_NUMBER:
                return
    except Exception as err:
        message = "update_layers_wms: {0}".format(
            err
        )
        check = Check(
            content_object=service,
            success=False,
            response_time=0,
            message=message
        )
        check.save()


def update_layers_wmts(service):
    """
    Update layers for an OGC:WMTS service.
    Sample endpoint: http://map1.vis.earthdata.nasa.gov/wmts-geo/1.0.0/WMTSCapabilities.xml
    """
    try:
        wmts = WebMapTileService(service.url)

        # set srs
        # WMTS is always in 4326
        srs, created = SpatialReferenceSystem.objects.get_or_create(code='EPSG:4326')
        service.srs.add(srs)

        layer_names = list(wmts.contents)
        layer_n = 0
        total = len(layer_names)
        for layer_name in layer_names:
            ows_layer = wmts.contents[layer_name]
            print 'Updating layer %s' % ows_layer.name
            layer, created = Layer.objects.get_or_create(name=ows_layer.name, service=service, catalog=service.catalog)
            if layer.active:
                links = [['OGC:WMTS', service.url]]
                layer.type = 'OGC:WMTS'
                layer.title = ows_layer.title
                layer.abstract = ows_layer.abstract
                # keywords
                # @tomkralidis wmts does not seem to support this attribute
                keywords = None
                if hasattr(ows_layer, 'keywords'):
                    keywords = ows_layer.keywords
                    for keyword in keywords:
                        layer.keywords.add(keyword)
                layer.url = service.url
                layer.page_url = layer.get_absolute_url
                links.append([
                    'WWW:LINK',
                    settings.SITE_URL.rstrip('/') + layer.page_url
                ])
                bbox = list(ows_layer.boundingBoxWGS84 or (-179.0, -89.0, 179.0, 89.0))
                layer.bbox_x0 = bbox[0]
                layer.bbox_y0 = bbox[1]
                layer.bbox_x1 = bbox[2]
                layer.bbox_y1 = bbox[3]
                layer.wkt_geometry = bbox2wktpolygon(bbox)
                layer.xml = create_metadata_record(
                    identifier=layer.id_string,
                    source=service.url,
                    links=links,
                    format='OGC:WMS',
                    type=layer.csw_type,
                    relation=service.id_string,
                    title=ows_layer.title,
                    alternative=ows_layer.name,
                    abstract=layer.abstract,
                    keywords=keywords,
                    wkt_geometry=layer.wkt_geometry
                )
                layer.anytext = gen_anytext(layer.title, layer.abstract, keywords)
                layer.save()
                # dates
                add_mined_dates(layer)
            layer_n = layer_n + 1
            # exits if DEBUG_SERVICES
            print "Updating layer n. %s/%s" % (layer_n, total)
            if DEBUG_SERVICES and layer_n == DEBUG_LAYER_NUMBER:
                return
    except Exception as err:
        message = "update_layers_wmts: {0}".format(
            err
        )
        check = Check(
            content_object=service,
            success=False,
            response_time=0,
            message=message
        )
        check.save()


def update_layers_wm(service):
    """
    Update layers for an WorldMap.
    Sample endpoint: http://worldmap.harvard.edu/
    """

    try:
        response = requests.get('http://worldmap.harvard.edu/data/search/api?start=0&limit=10')
        data = json.loads(response.content)
        total = data['total']

        # set srs
        # WorldMap supports only 4326, 900913, 3857
        for crs_code in ['EPSG:4326', 'EPSG:900913', 'EPSG:3857']:
            srs, created = SpatialReferenceSystem.objects.get_or_create(code=crs_code)
            service.srs.add(srs)

        layer_n = 0
        for i in range(0, total, 10):
            url = 'http://worldmap.harvard.edu/data/search/api?start=%s&limit=10' % i
            print 'Fetching %s' % url
            response = requests.get(url)
            data = json.loads(response.content)
            for row in data['rows']:
                name = row['name']
                title = row['title']
                abstract = row['abstract']
                bbox = row['bbox']
                page_url = row['detail']
                category = ''
                if 'topic_category' in row:
                    category = row['topic_category']
                username = ''
                if 'owner_username' in row:
                    username = row['owner_username']
                temporal_extent_start = ''
                if 'temporal_extent_start' in row:
                    temporal_extent_start = row['temporal_extent_start']
                temporal_extent_end = ''
                if 'temporal_extent_end' in row:
                    temporal_extent_end = row['temporal_extent_end']
                # we use the geoserver virtual layer getcapabilities for wm endpoint
                endpoint = 'http://worldmap.harvard.edu/geoserver/geonode/%s/wms?' % name
                is_public = True
                if '_permissions' in row:
                    if not row['_permissions']['view']:
                        is_public = False
                layer, created = Layer.objects.get_or_create(name=name, service=service, catalog=service.catalog)
                if layer.active:
                    links = [['Hypermap:WorldMap', endpoint]]
                    # update fields
                    layer.type = 'Hypermap:WorldMap'
                    layer.title = title
                    layer.abstract = abstract
                    layer.is_public = is_public
                    layer.url = endpoint
                    layer.page_url = page_url
                    # category and owner username
                    layer_wm, created = LayerWM.objects.get_or_create(layer=layer)
                    layer_wm.category = category
                    layer_wm.username = username
                    layer_wm.temporal_extent_start = temporal_extent_start
                    layer_wm.temporal_extent_end = temporal_extent_end
                    layer_wm.save()
                    # bbox
                    x0 = format_float(bbox['minx'])
                    y0 = format_float(bbox['miny'])
                    x1 = format_float(bbox['maxx'])
                    y1 = format_float(bbox['maxy'])
                    # In many cases for some reason to be fixed GeoServer has x coordinates flipped in WM.
                    x0, x1 = flip_coordinates(x0, x1)
                    y0, y1 = flip_coordinates(y0, y1)
                    layer.bbox_x0 = x0
                    layer.bbox_y0 = y0
                    layer.bbox_x1 = x1
                    layer.bbox_y1 = y1
                    # keywords
                    for keyword in row['keywords']:
                        layer.keywords.add(keyword)

                    layer.wkt_geometry = bbox2wktpolygon((bbox['minx'], bbox['miny'], bbox['maxx'], bbox['maxy']))
                    layer.xml = create_metadata_record(
                        identifier=layer.id_string,
                        source=endpoint,
                        links=links,
                        format='Hypermap:WorldMap',
                        type=layer.csw_type,
                        relation=service.id_string,
                        title=layer.title,
                        alternative=name,
                        abstract=layer.abstract,
                        keywords=row['keywords'],
                        wkt_geometry=layer.wkt_geometry
                    )
                    layer.anytext = gen_anytext(layer.title, layer.abstract, row['keywords'])
                    layer.save()
                    # dates
                    add_mined_dates(layer)
                    add_metadata_dates_to_layer([layer_wm.temporal_extent_start, layer_wm.temporal_extent_end], layer)
                layer_n = layer_n + 1
                # exits if DEBUG_SERVICES
                print "Updating layer n. %s/%s" % (layer_n, total)
                if DEBUG_SERVICES and layer_n == DEBUG_LAYER_NUMBER:
                    return
    except Exception as err:
        message = "update_layers_wm: {0}".format(
            err
        )
        check = Check(
            content_object=service,
            success=False,
            response_time=0,
            message=message
        )
        check.save()


def update_layers_warper(service):
    """
    Update layers for a Warper service.
    Sample endpoint: http://warp.worldmap.harvard.edu/maps
    """
    params = {'field': 'title', 'query': '', 'show_warped': '1', 'format': 'json'}
    headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
    request = requests.get(service.url, headers=headers, params=params)

    try:
        records = json.loads(request.content)
        total_pages = int(records['total_pages'])

        # set srs
        # Warper supports only 4326, 900913, 3857
        for crs_code in ['EPSG:4326', 'EPSG:900913', 'EPSG:3857']:
            srs, created = SpatialReferenceSystem.objects.get_or_create(code=crs_code)
            service.srs.add(srs)

        for i in range(1, total_pages + 1):
            params = {'field': 'title', 'query': '', 'show_warped': '1', 'format': 'json', 'page': i}
            request = requests.get(service.url, headers=headers, params=params)
            records = json.loads(request.content)
            print 'Fetched %s' % request.url
            layers = records['items']
            layer_n = 0
            total = len(layers)
            for layer in layers:
                name = layer['id']
                title = layer['title']
                abstract = layer['description']
                bbox = layer['bbox']
                # dates
                dates = []
                if 'published_date' in layer:
                    dates.append(layer['published_date'])
                if 'date_depicted' in layer:
                    dates.append(layer['date_depicted'])
                if 'depicts_year' in layer:
                    dates.append(layer['depicts_year'])
                if 'issue_year' in layer:
                    dates.append(layer['issue_year'])
                layer, created = Layer.objects.get_or_create(name=name, service=service, catalog=service.catalog)
                if layer.active:
                    # update fields
                    layer.type = 'Hypermap:WARPER'
                    layer.title = title
                    layer.abstract = abstract
                    layer.is_public = True
                    layer.url = '%s/wms/%s?' % (service.url, name)
                    layer.page_url = '%s/%s' % (service.url, name)
                    # bbox
                    x0 = None
                    y0 = None
                    x1 = None
                    y1 = None
                    if bbox:
                        bbox_list = bbox.split(',')
                        x0 = format_float(bbox_list[0])
                        y0 = format_float(bbox_list[1])
                        x1 = format_float(bbox_list[2])
                        y1 = format_float(bbox_list[3])
                    layer.bbox_x0 = x0
                    layer.bbox_y0 = y0
                    layer.bbox_x1 = x1
                    layer.bbox_y1 = y1
                    layer.save()
                    # dates
                    add_mined_dates(layer)
                    add_metadata_dates_to_layer(dates, layer)
                layer_n = layer_n + 1
                # exits if DEBUG_SERVICES
                print "Updating layer n. %s/%s" % (layer_n, total)
                if DEBUG_SERVICES and layer_n == DEBUG_LAYER_NUMBER:
                    return

    except Exception as err:
        message = "update_layers_warper: {0}. request={1} response={2}".format(
            err,
            service.url,
            request.text
        )
        check = Check(
            content_object=service,
            success=False,
            response_time=0,
            message=message
        )
        check.save()


def update_layers_esri_mapserver(service):
    """
    Update layers for an ESRI REST MapServer.
    Sample endpoint: https://gis.ngdc.noaa.gov/arcgis/rest/services/SampleWorldCities/MapServer/?f=json
    """
    try:
        esri_service = ArcMapService(service.url)
        # set srs
        # both mapserver and imageserver exposes just one srs at the service level
        # not sure if other ones are supported, for now we just store this one

        # not sure why this is needed, for now commenting out
        # if wkt_text:
        #     params = {'exact': 'True', 'error': 'True', 'mode': 'wkt', 'terms': wkt_text}
        #     req = requests.get('http://prj2epsg.org/search.json', params=params)
        #     object = json.loads(req.content)
        #     srs = int(object['codes'][0]['code'])

        srs_code = esri_service.spatialReference.wkid
        srs, created = SpatialReferenceSystem.objects.get_or_create(code=srs_code)
        service.srs.add(srs)
        # check if it has a WMS interface
        if 'supportedExtensions' in esri_service._json_struct:
            if 'WMSServer' in esri_service._json_struct['supportedExtensions']:
                # we need to change the url
                # http://cga1.cga.harvard.edu/arcgis/rest/services/ecuador/ecuadordata/MapServer?f=pjson
                # http://cga1.cga.harvard.edu/arcgis/services/ecuador/
                # ecuadordata/MapServer/WMSServer?request=GetCapabilities&service=WMS
                wms_url = service.url.replace('/rest/services/', '/services/')
                if '?f=pjson' in wms_url:
                    wms_url = wms_url.replace('?f=pjson', 'WMSServer?')
                if '?f=json' in wms_url:
                    wms_url = wms_url.replace('?f=json', 'WMSServer?')
                print 'This ESRI REST endpoint has an WMS interface to process: %s' % wms_url
                # import here as otherwise is circular (TODO refactor)
                from utils import create_service_from_endpoint
                create_service_from_endpoint(wms_url, 'OGC:WMS', catalog=service.catalog)
        # now process the REST interface
        layer_n = 0
        total = len(esri_service.layers)
        for esri_layer in esri_service.layers:
            # in some case the json is invalid
            # esri_layer._json_struct
            # {u'currentVersion': 10.01,
            # u'error':
            # {u'message': u'An unexpected error occurred processing the request.', u'code': 500, u'details': []}}
            if 'error' not in esri_layer._json_struct:
                print 'Updating layer %s' % esri_layer.name
                layer, created = Layer.objects.get_or_create(
                    name=esri_layer.id,
                    service=service,
                    catalog=service.catalog
                )
                if layer.active:
                    layer.type = 'ESRI:ArcGIS:MapServer'
                    links = [[layer.type, service.url]]
                    layer.title = esri_layer.name
                    layer.abstract = esri_service.serviceDescription
                    layer.url = service.url
                    layer.page_url = layer.get_absolute_url
                    links.append([
                        'WWW:LINK',
                        settings.SITE_URL.rstrip('/') + layer.page_url
                    ])
                    try:
                        layer.bbox_x0 = esri_layer.extent.xmin
                        layer.bbox_y0 = esri_layer.extent.ymin
                        layer.bbox_x1 = esri_layer.extent.xmax
                        layer.bbox_y1 = esri_layer.extent.ymax
                    except KeyError:
                        pass
                    try:
                        layer.bbox_x0 = esri_layer._json_struct['extent']['xmin']
                        layer.bbox_y0 = esri_layer._json_struct['extent']['ymin']
                        layer.bbox_x1 = esri_layer._json_struct['extent']['xmax']
                        layer.bbox_y1 = esri_layer._json_struct['extent']['ymax']
                    except Exception:
                        pass
                    layer.wkt_geometry = bbox2wktpolygon([layer.bbox_x0, layer.bbox_y0, layer.bbox_x1, layer.bbox_y1])
                    layer.xml = create_metadata_record(
                        identifier=layer.id_string,
                        source=service.url,
                        links=links,
                        format='ESRI:ArcGIS:MapServer',
                        type=layer.csw_type,
                        relation=service.id_string,
                        title=layer.title,
                        alternative=layer.title,
                        abstract=layer.abstract,
                        wkt_geometry=layer.wkt_geometry
                    )
                    layer.anytext = gen_anytext(layer.title, layer.abstract)
                    layer.save()
                    # dates
                    add_mined_dates(layer)
                layer_n = layer_n + 1
                # exits if DEBUG_SERVICES
                print "Updating layer n. %s/%s" % (layer_n, total)
                if DEBUG_SERVICES and layer_n == DEBUG_LAYER_NUMBER:
                    return
    except Exception as err:
        message = "update_layers_esri_mapserver: {0}".format(
            err
        )
        check = Check(
            content_object=service,
            success=False,
            response_time=0,
            message=message
        )
        check.save()


def update_layers_esri_imageserver(service):
    """
    Update layers for an ESRI REST ImageServer.
    Sample endpoint: https://gis.ngdc.noaa.gov/arcgis/rest/services/bag_bathymetry/ImageServer/?f=json
    """
    try:
        esri_service = ArcImageService(service.url)
        # set srs
        # both mapserver and imageserver exposes just one srs at the service level
        # not sure if other ones are supported, for now we just store this one
        obj = json.loads(esri_service._contents)
        srs_code = obj['spatialReference']['wkid']
        srs, created = SpatialReferenceSystem.objects.get_or_create(code=srs_code)
        service.srs.add(srs)
        layer, created = Layer.objects.get_or_create(name=obj['name'], service=service, catalog=service.catalog)
        if layer.active:
            layer.type = 'ESRI:ArcGIS:ImageServer'
            links = [[layer.type, service.url]]
            layer.title = obj['name']
            layer.abstract = esri_service.serviceDescription
            layer.url = service.url
            layer.bbox_x0 = str(obj['extent']['xmin'])
            layer.bbox_y0 = str(obj['extent']['ymin'])
            layer.bbox_x1 = str(obj['extent']['xmax'])
            layer.bbox_y1 = str(obj['extent']['ymax'])
            layer.page_url = layer.get_absolute_url
            links.append([
                'WWW:LINK',
                settings.SITE_URL.rstrip('/') + layer.page_url
            ])
            layer.wkt_geometry = bbox2wktpolygon([layer.bbox_x0, layer.bbox_y0, layer.bbox_x1, layer.bbox_y1])
            layer.xml = create_metadata_record(
                identifier=layer.id_string,
                source=service.url,
                links=links,
                format='ESRI:ArcGIS:ImageServer',
                type=layer.csw_type,
                relation=service.id_string,
                title=layer.title,
                alternative=layer.title,
                abstract=layer.abstract,
                wkt_geometry=layer.wkt_geometry
            )
            layer.anytext = gen_anytext(layer.title, layer.abstract)
            layer.save()
            # dates
            add_mined_dates(layer)
    except Exception as err:
        message = "update_layers_esri_imageserver: {0}".format(
            err
        )
        check = Check(
            content_object=service,
            success=False,
            response_time=0,
            message=message
        )
        check.save()


# signals

def endpointlist_post_save(instance, *args, **kwargs):
    """
    Used to process the lines of the endpoint list.
    """
    with open(instance.upload.file.name, mode='rb') as f:
        lines = f.readlines()
    for url in lines:
        if len(url) > 255:
            print 'Skipping this enpoint, as it is more than 255 characters: %s' % url
        else:
            if Endpoint.objects.filter(url=url, catalog=instance.catalog).count() == 0:
                endpoint = Endpoint(url=url, endpoint_list=instance)
                endpoint.catalog = instance.catalog
                endpoint.save()
    if not settings.REGISTRY_SKIP_CELERY:
        update_endpoints.delay(instance)
    else:
        update_endpoints(instance)


def endpoint_post_save(instance, *args, **kwargs):

    if Endpoint.objects.filter(url=instance.url).count() == 0:
        signals.post_save.disconnect(endpoint_post_save, sender=Endpoint)
        endpoint = Endpoint(url=instance.url)
        endpoint.save()
        signals.post_save.connect(endpoint_post_save, sender=Endpoint)
    if not settings.REGISTRY_SKIP_CELERY:
        update_endpoint.delay(instance)
    else:
        update_endpoint(instance)


def service_pre_save(instance, *args, **kwargs):
    """
    Used to do a service full check when saving it.
    """
    # check if service is unique
    # we cannot use unique_together as it relies on a combination of fields
    # from different models (service, resource)
    if Service.objects.filter(url=instance.url,
                              type=instance.type,
                              catalog=instance.catalog).count() > 0:
        raise Exception("There is already such a service. url={0} catalog={1}".format(
            instance.url, instance.catalog
        ))


def service_post_save(instance, *args, **kwargs):
    """
    Used to do a service full check when saving it.
    """
    # check service
    if not settings.REGISTRY_SKIP_CELERY:
        check_service.delay(instance)
    else:
        check_service(instance)


def layer_post_save(instance, *args, **kwargs):
    """
    Used to do a layer full check when saving it.
    """

    if instance.is_monitored:  # index and monitor
        if not settings.SKIP_CELERY_TASK:
            check_layer.delay(instance)
        else:
            check_layer(instance)
    else:  # just index
        check_layer(instance)


signals.post_save.connect(endpoint_post_save, sender=Endpoint)
signals.post_save.connect(endpointlist_post_save, sender=EndpointList)
signals.pre_save.connect(service_pre_save, sender=Service)
signals.post_save.connect(service_post_save, sender=Service)
signals.post_save.connect(layer_post_save, sender=Layer)



# -*- coding: utf-8 -*-
import requests
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from hypermap.aggregator.models import Catalog
from django.conf import settings
from .utils import parse_geo_box, request_time_facet, \
                request_heatmap_facet, gap_to_elastic, \
                asterisk_to_min_max
from .serializers import SearchSerializer, CatalogSerializer
import json

# - OPEN API specs
# https://github.com/OAI/OpenAPI-Specification/blob/master/versions/1.2.md#parameterObject

TIME_FILTER_FIELD = "layer_date"
GEO_FILTER_FIELD = "bbox"
GEO_HEATMAP_FIELD = "bbox"
USER_FIELD = "layer_originator"
TEXT_FIELD = "title"
TIME_SORT_FIELD = "layer_date"
GEO_SORT_FIELD = "bbox"

REGISTRY_SEARCH_URL = getattr(settings, "REGISTRY_SEARCH_URL", "elasticsearch+http://localhost:9200")

SEARCH_TYPE = REGISTRY_SEARCH_URL.split('+')[0]
SEARCH_URL = REGISTRY_SEARCH_URL.split('+')[1]


def elasticsearch(serializer, catalog):
    """
    https://www.elastic.co/guide/en/elasticsearch/reference/current/_the_search_api.html
    :param serializer:
    :return:
    """

    # search_engine_endpoint = serializer.validated_data.get("search_engine_endpoint")
    search_engine_endpoint = "{0}{1}/_search".format(SEARCH_URL, catalog.slug)

    q_text = serializer.validated_data.get("q_text")
    q_time = serializer.validated_data.get("q_time")
    q_geo = serializer.validated_data.get("q_geo")
    q_user = serializer.validated_data.get("q_user")
    d_docs_sort = serializer.validated_data.get("d_docs_sort")
    d_docs_limit = int(serializer.validated_data.get("d_docs_limit"))
    d_docs_page = int(serializer.validated_data.get("d_docs_page"))
    a_text_limit = serializer.validated_data.get("a_text_limit")
    a_user_limit = serializer.validated_data.get("a_user_limit")
    a_time_gap = serializer.validated_data.get("a_time_gap")
    a_time_limit = serializer.validated_data.get("a_time_limit")
    original_response = serializer.validated_data.get("original_response")

    # Dict for search on Elastic engine
    must_array = []
    filter_dic = {}
    aggs_dic = {}

    # String searching
    if q_text:
        query_string = {
            "query_string": {
                "query": q_text
            }
        }
        # add string searching
        must_array.append(query_string)

    if q_time:
        # check if q_time exists
        q_time = str(q_time)  # check string
        shortener = q_time[1:-1]
        shortener = shortener.split(" TO ")
        gte = shortener[0]  # greater than
        lte = shortener[1]  # less than
        layer_date = {}
        if gte == '*' and lte != '*':
            layer_date["lte"] = lte
            range_time = {
                "layer_date": layer_date
            }
            range_time = {"range": range_time}
            must_array.append(range_time)
        if gte != '*' and lte == '*':
            layer_date["gte"] = gte
            range_time = {
                "layer_date": layer_date
            }
            range_time = {"range": range_time}
            must_array.append(range_time)
        if gte != '*' and lte != '*':
            layer_date["gte"] = gte
            layer_date["lte"] = lte
            range_time = {
                "layer_date": layer_date
            }
            range_time = {"range": range_time}
            must_array.append(range_time)
    # geo_shape searching
    if q_geo:
        q_geo = str(q_geo)
        q_geo = q_geo[1:-1]
        Ymin, Xmin = q_geo.split(" TO ")[0].split(",")
        Ymax, Xmax = q_geo.split(" TO ")[1].split(",")
        geoshape_query = {
            "layer_geoshape": {
                "shape": {
                    "type": "envelope",
                    "coordinates": [[Xmin, Ymax], [Xmax, Ymin]]
                },
                "relation": "within"
            }
        }
        filter_dic["geo_shape"] = geoshape_query

    if q_user:
        # Using q_user
        user_searching = {
            "match": {
                "layer_originator": q_user
            }
        }
        must_array.append(user_searching)

    dic_query = {
        "query": {
            "bool": {
                "must": must_array,
                "filter": filter_dic
            }
        }
    }
    # Page
    if d_docs_limit:
        dic_query["size"] = d_docs_limit

    if d_docs_page:
        dic_query["from"] = d_docs_limit * d_docs_page - d_docs_limit

    if d_docs_sort == "score":
        dic_query["sort"] = {"_score": {"order": "desc"}}

    if d_docs_sort == "time":
        dic_query["sort"] = {"layer_date": {"order": "desc"}}

    if d_docs_sort == "distance":
        if q_geo:
            # distance_x = float(((float(Xmin) - float(Xmax)) ** 2.0) ** (0.5))
            # distance_y = float(((float(Ymin) - float(Ymax)) ** 2.0) ** (0.5))
            msg = ("Sorting by distance is different on ElasticSearch than Solr, because this"
                   "feature on elastic is unavailable to geo_shape type.ElasticSearch docs said:"
                   "Due to the complex input structure and index representation of shapes,"
                   "it is not currently possible to sort shapes or retrieve their fields directly."
                   "The geo_shape value is only retrievable through the _source field."
                   " Link: https://www.elastic.co/guide/en/elasticsearch/reference/current/geo-shape.html")
            return {"error": {"msg": msg}}

        else:
            msg = "q_qeo MUST BE NO ZERO if you wanna sort by distance"
            return {"error": {"msg": msg}}

    if a_text_limit:
        # getting most frequently occurring users.
        text_limit = {
            "terms": {
                "field": "abstract",
                "size": a_text_limit
            }
        }
        aggs_dic['popular_text'] = text_limit

    if a_user_limit:
        # getting most frequently occurring users.
        users_limit = {

            "terms": {
                "field": "layer_originator",
                "size": a_user_limit
            }
        }
        aggs_dic['popular_users'] = users_limit

    if a_time_limit:
        # Work in progress
        if q_time:
            if not a_time_gap:
                # getting time limit histogram.
                time_limt = {
                    "date_range": {
                        "field": "layer_date",
                        "format": "yyyy-MM-dd'T'HH:mm:ssZ",
                        "ranges": [
                            {"from": gte, "to": lte}
                        ]
                    }
                }
                aggs_dic['range'] = time_limt
            else:
                pass

        else:
            msg = "If you want to use a_time_limit feature, q_time MUST BE initialized"
            return {"error": {"msg": msg}}

    if a_time_gap:
        interval = gap_to_elastic(a_time_gap)
        time_gap = {
            "date_histogram": {
                "field": "layer_date",
                "format": "yyyy-MM-dd'T'HH:mm:ssZ",
                "interval": interval
            }
        }
        aggs_dic['articles_over_time'] = time_gap

    # adding aggreations on body query
    if aggs_dic:
        dic_query['aggs'] = aggs_dic
    try:
        res = requests.post(search_engine_endpoint, data=json.dumps(dic_query))
    except Exception as e:
        return 500, {"error": {"msg": str(e)}}

    es_response = res.json()

    if original_response:
        return es_response

    data = {}

    if 'error' in es_response:
        data["error"] = es_response["error"]
        return 400, data

    data["request_url"] = res.url
    data["request_body"] = json.dumps(dic_query)
    data["a.matchDocs"] = es_response['hits']['total']
    docs = []
    # aggreations response: facets searching
    if 'aggregations' in es_response:
        aggs = es_response['aggregations']
        # getting the most frequently occurring users.
        if 'popular_users' in aggs:
            a_users_list_array = []
            users_resp = aggs["popular_users"]["buckets"]
            for item in users_resp:
                temp = {}
                temp['count'] = item['doc_count']
                temp['value'] = item['key']
                a_users_list_array.append(temp)
            data["a.user"] = a_users_list_array

        # getting most frequently ocurring words
        if 'popular_text' in aggs:
            a_text_list_array = []
            text_resp = es_response["aggregations"]["popular_text"]["buckets"]
            for item in text_resp:
                temp = {}
                temp['count'] = item['doc_count']
                temp['value'] = item['key']
                a_text_list_array.append(temp)
            data["a.text"] = a_text_list_array

        if 'articles_over_time' in aggs:
            gap_count = []
            a_gap = {}
            gap_resp = aggs["articles_over_time"]["buckets"]

            start = "*"
            end = "*"

            if len(gap_resp) > 0:
                start = gap_resp[0]['key_as_string'].replace('+0000', 'z')
                end = gap_resp[-1]['key_as_string'].replace('+0000', 'z')

            a_gap['start'] = start
            a_gap['end'] = end
            a_gap['gap'] = a_time_gap

            for item in gap_resp:
                temp = {}
                if item['doc_count'] != 0:
                    temp['count'] = item['doc_count']
                    temp['value'] = item['key_as_string'].replace('+0000', 'z')
                    gap_count.append(temp)
            a_gap['counts'] = gap_count
            data['a.time'] = a_gap

        if 'range' in aggs:
            # Work in progress
            # Pay attention in the following code lines: Make it better!!!!
            time_count = []
            time_resp = aggs["range"]["buckets"]
            a_time = {}
            a_time['start'] = gte
            a_time['end'] = lte
            a_time['gap'] = None

            for item in time_resp:
                temp = {}
                if item['doc_count'] != 0:
                    temp['count'] = item['doc_count']
                    temp['value'] = item['key'].replace('+0000', 'z')
                    time_count.append(temp)
            a_time['counts'] = time_count
            data['a.time'] = a_time

    if not int(d_docs_limit) == 0:
        for item in es_response['hits']['hits']:
            # data
            temp = item['_source']['abstract']
            temp = temp.replace(u'\u201c', "\"")
            temp = temp.replace(u'\u201d', "\"")
            temp = temp.replace('"', "\"")
            temp = temp.replace("'", "\'")
            temp = temp.replace(u'\u2019', "\'")
            item['_source']['abstract'] = temp
            docs.append(item['_source'])

    data["d.docs"] = docs

    return data


def solr(serializer):
    """
    Search on solr endpoint
    :param serializer:
    :return:
    """
    search_engine_endpoint = serializer.validated_data.get("search_engine_endpoint")
    q_time = serializer.validated_data.get("q_time")
    q_geo = serializer.validated_data.get("q_geo")
    q_text = serializer.validated_data.get("q_text")
    q_user = serializer.validated_data.get("q_user")
    d_docs_limit = serializer.validated_data.get("d_docs_limit")
    d_docs_page = serializer.validated_data.get("d_docs_page")
    d_docs_sort = serializer.validated_data.get("d_docs_sort")
    a_time_limit = serializer.validated_data.get("a_time_limit")
    a_time_gap = serializer.validated_data.get("a_time_gap")
    a_time_filter = serializer.validated_data.get("a_time_filter")
    a_hm_limit = serializer.validated_data.get("a_hm_limit")
    a_hm_gridlevel = serializer.validated_data.get("a_hm_gridlevel")
    a_hm_filter = serializer.validated_data.get("a_hm_filter")
    a_text_limit = serializer.validated_data.get("a_text_limit")
    a_user_limit = serializer.validated_data.get("a_user_limit")
    original_response = serializer.validated_data.get("original_response")

    # query params to be sent via restful solr
    params = {
        "q": "*:*",
        "indent": "on",
        "wt": "json",
        "rows": d_docs_limit,
        "facet": "off",
        "facet.field": [],
        "debug": "timing"
    }
    if q_text:
        params["q"] = q_text

    if d_docs_limit >= 0:
        d_docs_page -= 1
        d_docs_page = d_docs_limit * d_docs_page
        params["start"] = d_docs_page

    # query params for filters
    filters = []
    if q_time:
        # TODO: when user sends incomplete dates like 2000, its completed: 2000-(TODAY-MONTH)-(TODAY-DAY)T00:00:00Z
        # TODO: "Invalid Date in Date Math String:'[* TO 2000-12-05T00:00:00Z]'"
        # Kotlin like: "{!field f=layer_date tag=layer_date}[* TO 2000-12-05T00:00:00Z]"
        # then do it simple:
        filters.append("{0}:{1}".format(TIME_FILTER_FIELD, q_time))
    if q_geo:
        filters.append("{0}:{1}".format(GEO_FILTER_FIELD, q_geo))

    if q_user:
        filters.append("{{!field f={0} tag={0}}}{1}".format(USER_FIELD, q_user))

    if filters:
        params["fq"] = filters

    # query params for ordering
    if d_docs_sort == 'score' and q_text:
        params["sort"] = 'score desc'
    elif d_docs_sort == 'time':
        params["sort"] = '{} desc'.format(TIME_SORT_FIELD)
    elif d_docs_sort == 'distance':
        rectangle = parse_geo_box(q_geo)
        params["sort"] = 'geodist() asc'
        params["sfield"] = GEO_SORT_FIELD
        params["pt"] = '{0},{1}'.format(rectangle.centroid.x, rectangle.centroid.y)

    # query params for facets
    if a_time_limit > 0:
        params["facet"] = 'on'
        time_filter = a_time_filter or q_time or None

        # traduce * to actual min/max dates.
        time_filter = asterisk_to_min_max(TIME_FILTER_FIELD, time_filter, search_engine_endpoint)

        # create the range faceting params.
        facet_parms = request_time_facet(TIME_FILTER_FIELD, time_filter, a_time_gap, a_time_limit)
        params.update(facet_parms)

    if a_hm_limit > 0:
        params["facet"] = 'on'
        hm_facet_params = request_heatmap_facet(GEO_HEATMAP_FIELD, a_hm_filter, a_hm_gridlevel, a_hm_limit)
        params.update(hm_facet_params)

    if a_text_limit > 0:
        params["facet"] = 'on'
        params["facet.field"].append(TEXT_FIELD)
        params["f.{}.facet.limit".format(TEXT_FIELD)] = a_text_limit

    if a_user_limit > 0:
        params["facet"] = 'on'
        params["facet.field"].append("{{! ex={0}}}{0}".format(USER_FIELD))
        params["f.{}.facet.limit".format(USER_FIELD)] = a_user_limit

    try:
        res = requests.get(
            search_engine_endpoint, params=params
        )
    except Exception as e:
        return 500, {"error": {"msg": str(e)}}

    print '>', res.url

    solr_response = res.json()
    solr_response["solr_request"] = res.url

    if original_response > 0:
        return solr_response

    # create the response dict following the swagger model:
    data = {}

    if 'error' in solr_response:
        data["error"] = solr_response["error"]
        return 400, data

    response = solr_response["response"]
    data["a.matchDocs"] = response.get("numFound")

    if response.get("docs"):
        data["d.docs"] = response.get("docs")

    if a_time_limit > 0:
        date_facet = solr_response["facet_counts"]["facet_ranges"][TIME_FILTER_FIELD]
        counts = []
        value_count = iter(date_facet.get("counts"))
        for value, count in zip(value_count, value_count):
            counts.append({
                "value": value,
                "count": count
            })
        a_time = {
            "start": date_facet.get("start"),
            "end": date_facet.get("end"),
            "gap": date_facet.get("gap"),
            "counts": counts
        }
        data["a.time"] = a_time

    if a_hm_limit > 0:
        hm_facet_raw = solr_response["facet_counts"]["facet_heatmaps"][GEO_HEATMAP_FIELD]
        hm_facet = {
            'gridLevel': hm_facet_raw[1],
            'columns': hm_facet_raw[3],
            'rows': hm_facet_raw[5],
            'minX': hm_facet_raw[7],
            'maxX': hm_facet_raw[9],
            'minY': hm_facet_raw[11],
            'maxY': hm_facet_raw[13],
            'counts_ints2D': hm_facet_raw[15],
            'projection': 'EPSG:4326'
        }
        data["a.hm"] = hm_facet

    if a_user_limit > 0:
        user_facet = solr_response["facet_counts"]["facet_fields"][USER_FIELD]

        counts = []
        value_count = iter(user_facet)
        for value, count in zip(value_count, value_count):
            counts.append({
                "value": value,
                "count": count
            })
        data["a.user"] = counts

    if a_text_limit > 0:
        text_facet = solr_response["facet_counts"]["facet_fields"][TEXT_FIELD]

        counts = []
        value_count = iter(text_facet)
        for value, count in zip(value_count, value_count):
            counts.append({
                "value": value,
                "count": count
            })
        data["a.text"] = counts

    subs = []
    for label, values in solr_response["debug"]["timing"].iteritems():
        if type(values) is not dict:
            continue
        subs_data = {"label": label, "subs": []}
        for label, values in values.iteritems():
            if type(values) is not dict:
                subs_data["millis"] = values
                continue
            subs_data["subs"].append({
                "label": label,
                "millis": values.get("time")
            })
        subs.append(subs_data)

    timing = {
        "label": "requests.get.elapsed",
        "millis": res.elapsed,
        "subs": [{
            "label": "QTime",
            "millis": solr_response["responseHeader"].get("QTime"),
            "subs": subs
        }]
    }

    data["timing"] = timing
    data["request_url"] = res.url

    return data


class Search(APIView):
    """
    Swagger docs located in hypermap/api/static/swagger.yaml
    edit in http://editor.swagger.io/#/
    """

    def get(self, request, catalog_slug):

        serializer = SearchSerializer(data=request.GET)
        if serializer.is_valid(raise_exception=True):

            try:
                catalog = Catalog.objects.get(slug=catalog_slug)
            except Catalog.DoesNotExist:
                return Response({"error": "catalog '{}' not found".format(catalog_slug)},
                                status=404)

            # check if data source is remote
            # if catalog.is_remote and request.META['SERVER_PORT'] == "8000":
            if catalog.is_remote:
                response = requests.get(catalog.url, params=request.query_params)
                if response.status_code in [200, 400]:
                    return Response(response.json(),
                                    status=response.status_code)
                else:
                    return Response(response.text,
                                    status=response.status_code)

            search_engine = serializer.validated_data.get("search_engine", "elasticsearch")
            if search_engine == 'solr':
                data = solr(serializer)
            else:
                data = elasticsearch(serializer, catalog)

            status = 200
            if type(data) is tuple:
                status = data[0]
                data = data[1]

            return Response(data, status=status)


class CatalogViewSet(ModelViewSet):
    queryset = Catalog.objects.all()
    serializer_class = CatalogSerializer



from .default import *  # noqa
from .default import MIDDLEWARE_CLASSES
from .default import INSTALLED_APPS
import os


SITE_URL = 'http://192.168.33.15:8000/'

SEARCH_ENABLED = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': os.getenv('DATABASE_NAME', 'hypermap'),
        'USER': os.getenv('DATABASE_USER', 'hypermap'),
        'PASSWORD': os.getenv('DATABASE_PASSWORD', 'hypermap'),
        'HOST': os.getenv('DATABASE_HOST', '127.0.0.1'),
        'PORT': os.getenv('DATABASE_PORT', '5432'),
    }
}

# setup for developer toolbar
DEBUG_TOOLBAR_PATCH_SETTINGS = False

INSTALLED_APPS = INSTALLED_APPS + (
    'debug_toolbar',
)

MIDDLEWARE_CLASSES = MIDDLEWARE_CLASSES + (
    'debug_toolbar.middleware.DebugToolbarMiddleware',
)

INTERNAL_IPS = ('192.168.33.1',)

# Load more settings from a file called local_settings.py if it exists
try:
    from local_settings import *  # noqa
except ImportError:
    pass



from django.conf import settings
from django.conf.urls import include, patterns, url

from . import views

urlpatterns = [
    url(r'^(?P<catalog_slug>[\w-]+)/csw/$', views.csw_global_dispatch_by_catalog,
        name='csw_global_dispatch_by_catalog'),
    url(r'^opensearch$', views.opensearch_dispatch, name='opensearch_dispatch')
]

if settings.DEBUG:
    import debug_toolbar
    urlpatterns += patterns(
        '',
        url(r'^__debug__/', include(debug_toolbar.urls)),
    )



import urllib2
import logging
import requests
import re
import sys
import math
import traceback
from urlparse import urlparse

from django.conf import settings
from lxml import etree
from owslib.csw import CatalogueServiceWeb, CswRecord
from owslib.wms import WebMapService
from owslib.tms import TileMapService
from owslib.wmts import WebMapTileService
from arcrest import Folder as ArcFolder

from hypermap.aggregator.enums import SERVICE_TYPES
from lxml.etree import XMLSyntaxError

LOGGER = logging.getLogger(__name__)


def create_layer_from_metadata_xml(resourcetype, xml, monitor=False):
    """
    Create a layer / keyword list from a metadata record if it does not already exist.
    """
    from models import gen_anytext, Layer

    if resourcetype == 'http://www.opengis.net/cat/csw/2.0.2':  # Dublin core
        md = CswRecord(etree.fromstring(xml))

    layer = Layer(
        is_monitored=monitor,
        name=md.title,
        title=md.title,
        abstract=md.abstract,
        xml=xml,
        anytext=gen_anytext(md.title, md.abstract, md.subjects)
    )

    if md.bbox is not None:
        layer.bbox_x0 = format_float(md.bbox.minx)
        layer.bbox_y0 = format_float(md.bbox.miny)
        layer.bbox_x1 = format_float(md.bbox.maxx)
        layer.bbox_y1 = format_float(md.bbox.maxy)

        layer.wkt_geometry = bbox2wktpolygon([md.bbox.minx, md.bbox.miny, md.bbox.maxx, md.bbox.maxy])

    return layer, md.subjects


def create_service_from_endpoint(endpoint, service_type, title=None, abstract=None, catalog=None):
    """
    Create a service from an endpoint if it does not already exists.
    """
    from models import Service
    if Service.objects.filter(url=endpoint, catalog=catalog).count() == 0:
        # check if endpoint is valid
        request = requests.get(endpoint)
        if request.status_code == 200:
            print 'Creating a %s service for endpoint=%s catalog=%s' % (service_type, endpoint, catalog)
            service = Service(
                        type=service_type, url=endpoint, title=title, abstract=abstract,
                        csw_type='service', catalog=catalog
                        )
            service.save()
            return service
        else:
            print 'This endpoint is invalid, status code is %s' % request.status_code
    else:
        print 'A service for this endpoint %s in catalog %s already exists' % (endpoint, catalog)
        return None


def create_services_from_endpoint(url, catalog, greedy_opt=True):
    """
    Generate service/services from an endpoint.
    WMS, WMTS, TMS endpoints correspond to a single service.
    ESRI, CSW endpoints corrispond to many services.
    :return: imported, message
    """

    # this variable will collect any exception message during the routine.
    # will be used in the last step to send a message if "detected" var is False.
    messages = []

    num_created = 0
    endpoint = get_sanitized_endpoint(url)
    try:
        urllib2.urlopen(endpoint, timeout=10)
    except Exception as e:
        message = traceback.format_exception(*sys.exc_info())
        print 'ERROR! Cannot open this endpoint: %s' % endpoint
        print 'ERROR MESSAGE: %s' % message
        return False, message

    detected = False

    # handle specific service types for some domains (WorldMap, Wrapper...)
    parsed_uri = urlparse(endpoint)
    domain = '{uri.scheme}://{uri.netloc}/'.format(uri=parsed_uri)
    if domain == 'http://worldmap.harvard.edu/':
        service_type = 'Hypermap:WorldMap'
        title = 'Harvard WorldMap'
        abstract = 'Harvard WorldMap'
        endpoint = domain
        detected = True
    if domain in [
        'http://maps.nypl.org/',
        'http://mapwarper.net/',
        'http://warp.worldmap.harvard.edu/',
    ]:
        service_type = 'Hypermap:WARPER'
        title = 'Warper at %s' % domain
        abstract = 'Warper at %s' % domain
        detected = True

    # test if it is CSW, WMS, TMS, WMTS or Esri
    # CSW
    try:
        csw = CatalogueServiceWeb(endpoint)
        service_type = 'OGC:CSW'
        service_links = {}
        detected = True

        typenames = 'csw:Record'
        outputschema = 'http://www.opengis.net/cat/csw/2.0.2'

        if 'csw_harvest_pagesize' in settings.PYCSW['manager']:
            pagesize = int(settings.PYCSW['manager']['csw_harvest_pagesize'])
        else:
            pagesize = 10

        print 'Harvesting CSW %s' % endpoint
        # now get all records
        # get total number of records to loop against
        try:
            csw.getrecords2(typenames=typenames, resulttype='hits',
                            outputschema=outputschema)
            matches = csw.results['matches']
        except:  # this is a CSW, but server rejects query
            raise RuntimeError(csw.response)

        if pagesize > matches:
            pagesize = matches

        print 'Harvesting %d CSW records' % matches

        # loop over all catalogue records incrementally
        for r in range(1, matches+1, pagesize):
            try:
                csw.getrecords2(typenames=typenames, startposition=r,
                                maxrecords=pagesize, outputschema=outputschema, esn='full')
            except Exception as err:  # this is a CSW, but server rejects query
                raise RuntimeError(csw.response)
            for k, v in csw.records.items():
                # try to parse metadata
                try:
                    LOGGER.info('Looking for service links')
                    if v.references:  # not empty
                        for ref in v.references:
                            if ref['scheme'] in [st[0] for st in SERVICE_TYPES]:
                                if ref['url'] not in service_links:
                                    service_links[ref['url']] = ref['scheme']
                except Exception as err:  # parsing failed for some reason
                    LOGGER.warning('Metadata parsing failed %s', err)

        LOGGER.info('Service links found: %s', service_links)
        for k, v in service_links.items():
            try:
                service = create_service_from_endpoint(k, v, catalog=catalog)
                if service is not None:
                    num_created = num_created + 1
            except Exception as err:
                raise RuntimeError('HHypermap error: %s' % err)
    except XMLSyntaxError as e:
        # This is not XML, so likely not a CSW. Moving on.
        pass
    except Exception as e:
        print str(e)
        messages.append(str(e))

    # WMS
    if not detected:
        try:
            service = WebMapService(endpoint, timeout=10)
            service_type = 'OGC:WMS'
            title = service.identification.title,
            abstract = service.identification.abstract
            detected = True
        except XMLSyntaxError as e:
            # This is not XML, so likely not a WMS. Moving on.
            pass
        except Exception as e:
            print str(e)
            messages.append(str(e))

    # TMS
    if not detected:
        try:
            service = TileMapService(endpoint, timeout=10)
            service_type = 'OSGeo:TMS'
            title = service.identification.title,
            abstract = service.identification.abstract
            detected = True
        except XMLSyntaxError as e:
            # This is not XML, so likely not a TsMS. Moving on.
            pass
        except Exception as e:
            print str(e)
            messages.append(str(e))

    # WMTS
    if not detected:
        try:
            # @tomkralidis timeout is not implemented for WebMapTileService?
            service = WebMapTileService(endpoint)
            service_type = 'OGC:WMTS'
            title = service.identification.title,
            abstract = service.identification.abstract
            detected = True
        except XMLSyntaxError as e:
            # This is not XML, so likely not a WMTS. Moving on.
            pass
        except Exception as e:
            print str(e)
            messages.append(str(e))

    # if detected, let's create the service
    if detected and service_type != 'OGC:CSW':
        try:
            service = create_service_from_endpoint(
                endpoint,
                service_type,
                title,
                abstract=abstract,
                catalog=catalog
            )
            if service is not None:
                num_created = num_created + 1
        except XMLSyntaxError as e:
            # This is not XML, so likely not a OGC:CSW. Moving on.
            pass
        except Exception as e:
            print str(e)
            messages.append(str(e))

    # Esri
    # a good sample is here: https://gis.ngdc.noaa.gov/arcgis/rest/services

    # we can safely assume the following condition (at least it is true for 1170 services)
    # we need to test this as ArcFolder can freeze with not esri url such as this one:
    # http://hh.worldmap.harvard.edu/admin/aggregator/service/?q=%2Frest%2Fservices

    if '/rest/services' in endpoint:
        if not detected:
            try:
                esri = ArcFolder(endpoint)
                service_type = 'ESRI'
                detected = True
                only_one_sw = False
                services_folder = esri.folders

                def split_service(s):
                    return s.url.split(url_token)[1].split('/')

                root_services = process_esri_services(esri.services, catalog)
                num_created = num_created + len(root_services)

                # Enable the user to fetch a single service of a single folder.
                if not greedy_opt:
                    # Get folder and service from endpoint url.
                    url_token = '/rest/services/'
                    components = url.strip().split(url_token)[1].split('/')
                    esri_folder, esri_service = components[0], components[1]

                    single_folder = [f for f in esri.folders if split_service(f)[0] == esri_folder]
                    if len(single_folder) != 0:
                        only_one_sw = True
                        services_folder = single_folder

                for folder in services_folder:
                    services_to_process = folder.services
                    print services_to_process

                    # Get folder class from the list of esri services for the given endpoint.
                    if only_one_sw:
                        fs = services_folder[0].services
                        # Get service class from the folder for the given endpoint.
                        services_to_process = [s for s in fs if split_service(s)[1] == esri_service]

                    folder_services = process_esri_services(services_to_process, catalog)
                    num_created = num_created + len(folder_services)
            except Exception as e:
                print str(e)
                messages.append(str(e))

    if detected:
        return True, '%s service/s created' % num_created
    else:
        m = '|'.join(messages)
        return False, 'ERROR! Could not detect service type for ' \
                      'endpoint %s or already existing. messages=(%s)' % (endpoint, m)


def process_esri_services(esri_services, catalog):
    services_created = []
    for esri_service in esri_services:
        # for now we process only MapServer
        if '/MapServer/' in esri_service.url:
            # we import only MapServer with at least one layer
            if hasattr(esri_service, 'layers'):
                service = create_service_from_endpoint(
                    esri_service.url,
                    'ESRI:ArcGIS:MapServer',
                    esri_service.mapName,
                    esri_service.description,
                    catalog=catalog
                )
                services_created.append(service)

        # Don't process ImageServer until the following issue has been resolved:
        # https://github.com/mapproxy/mapproxy/issues/235
        # if '/ImageServer/' in esri_service.url:
        #     service = create_service_from_endpoint(
        #         esri_service.url,
        #         'ESRI:ArcGIS:ImageServer',
        #         '',
        #         esri_service.serviceDescription
        #     )
        #      services_created.append(service)

    return services_created


def inverse_mercator(xy):
    """
        Given coordinates in spherical mercator, return a lon,lat tuple.
    """
    lon = (xy[0] / 20037508.34) * 180
    lat = (xy[1] / 20037508.34) * 180
    lat = 180 / math.pi * \
        (2 * math.atan(math.exp(lat * math.pi / 180)) - math.pi / 2)
    return (lon, lat)


def mercator_to_llbbox(bbox):
    minlonlat = inverse_mercator([bbox[0], bbox[1]])
    maxlonlat = inverse_mercator([bbox[2], bbox[3]])
    return [minlonlat[0], minlonlat[1], maxlonlat[0], maxlonlat[1]]


def get_sanitized_endpoint(url):
    """
    Sanitize an endpoint, as removing unneeded parameters
    """
    # sanitize esri
    sanitized_url = url.rstrip()
    esri_string = '/rest/services'
    if esri_string in url:
        match = re.search(esri_string, sanitized_url)
        sanitized_url = url[0:(match.start(0)+len(esri_string))]
    return sanitized_url


def get_esri_service_name(url):
    """
    A method to get a service name from an esri endpoint.
    For example: http://example.com/arcgis/rest/services/myservice/mylayer/MapServer/?f=json
    Will return: myservice/mylayer
    """
    result = re.search('rest/services/(.*)/[MapServer|ImageServer]', url)
    if result is None:
        return url
    else:
        return result.group(1)


def get_esri_extent(esriobj):
    """
    Get the extent of an ESRI resource
    """

    extent = None
    srs = None

    if 'fullExtent' in esriobj._json_struct:
        extent = esriobj._json_struct['fullExtent']
    if 'extent' in esriobj._json_struct:
        extent = esriobj._json_struct['extent']

    try:
        srs = extent['spatialReference']['wkid']
    except KeyError as err:
        LOGGER.error(err)

    return [extent, srs]


def flip_coordinates(c1, c2):
    if c1 > c2:
        print 'Flipping coordinates %s, %s' % (c1, c2)
        temp = c1
        c1 = c2
        c2 = temp
    return c1, c2


def format_float(value):
    if value is None:
        return None
    try:
        value = float(value)
        if value > 999999999:
            return None
        return value
    except ValueError:
        return None


def bbox2wktpolygon(bbox):
    """
    Return OGC WKT Polygon of a simple bbox list of strings
    """

    minx = float(bbox[0])
    miny = float(bbox[1])
    maxx = float(bbox[2])
    maxy = float(bbox[3])
    return 'POLYGON((%.2f %.2f, %.2f %.2f, %.2f %.2f, %.2f %.2f, %.2f %.2f))' \
        % (minx, miny, minx, maxy, maxx, maxy, maxx, miny, minx, miny)



from .default import *  # noqa
from .default import BASE_DIR
import os

REGISTRY_SKIP_CELERY = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

SOLR_ENABLED = True
SOLR_URL = 'http://localhost:8983/solr/hypermap_test'
REGISTRY_SEARCH_URL = "elasticsearch+http://localhost:9200"

SEARCH_URL = SOLR_URL

# BROKER_BACKEND = 'memory'
# BROKER_URL='memory://'
# CELERY_ALWAYS_EAGER = True
# CELERY_EAGER_PROPAGATES_EXCEPTIONS = True
# TEST_RUNNER = 'djcelery.contrib.test_runner.CeleryTestSuiteRunner'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import taggit.managers
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Catalog',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(help_text=b'Display name in UI', max_length=255)),
                ('slug', django_extensions.db.fields.AutoSlugField(editable=False, populate_from=b'name', blank=True, help_text=b'Leave empty to be populated from name')),
                ('url', models.URLField(help_text=b'Only if remote. URL where the API for the search backend is served.ex: http://localhost:8000/registry/api/search/', max_length=255, null=True, blank=True)),
            ],
        ),
        migrations.CreateModel(
            name='Check',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('object_id', models.PositiveIntegerField()),
                ('checked_datetime', models.DateTimeField(auto_now=True)),
                ('success', models.BooleanField(default=False)),
                ('response_time', models.FloatField()),
                ('message', models.TextField(default=b'OK')),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='Endpoint',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('processed', models.BooleanField(default=False)),
                ('processed_datetime', models.DateTimeField(auto_now=True)),
                ('imported', models.BooleanField(default=False)),
                ('message', models.TextField(null=True, blank=True)),
                ('url', models.URLField(max_length=255)),
                ('catalog', models.ForeignKey(default=1, editable=False, to='aggregator.Catalog')),
            ],
        ),
        migrations.CreateModel(
            name='EndpointList',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('upload', models.FileField(upload_to=b'endpoint_lists')),
                ('greedy', models.BooleanField(default=False)),
                ('catalog', models.ForeignKey(default=1, editable=False, to='aggregator.Catalog')),
            ],
        ),
        migrations.CreateModel(
            name='Layer',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(max_length=255, null=True, blank=True)),
                ('abstract', models.TextField(null=True, blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_updated', models.DateTimeField(auto_now=True)),
                ('active', models.BooleanField(default=True)),
                ('url', models.URLField(max_length=255)),
                ('is_public', models.BooleanField(default=True)),
                ('type', models.CharField(max_length=32, choices=[(b'OGC:CSW', b'Catalogue Service for the Web (CSW)'), (b'OGC:WMS', b'Web Map Service (WMS)'), (b'OGC:WMTS', b'Web Map Tile Service (WMTS)'), (b'OSGeo:TMS', b'Tile Map Service (TMS)'), (b'ESRI:ArcGIS:MapServer', b'ArcGIS REST MapServer'), (b'ESRI:ArcGIS:ImageServer', b'ArcGIS REST ImageServer'), (b'Hypermap:WorldMap', b'Harvard WorldMap'), (b'Hypermap:WARPER', b'Mapwarper')])),
                ('temporal_extent_start', models.CharField(max_length=255, null=True, blank=True)),
                ('temporal_extent_end', models.CharField(max_length=255, null=True, blank=True)),
                ('csw_last_updated', models.CharField(default=b'2016-08-15T20:00:11Z', max_length=32, null=True, blank=True)),
                ('csw_type', models.CharField(default=b'dataset', max_length=32)),
                ('csw_typename', models.CharField(default=b'csw:Record', max_length=32)),
                ('csw_schema', models.CharField(default=b'http://www.opengis.net/cat/csw/2.0.2', max_length=64)),
                ('anytext', models.TextField(null=True, blank=True)),
                ('wkt_geometry', models.TextField(default=b'POLYGON((-180 -90,-180 90,180 90,180 -90,-180 -90))')),
                ('xml', models.TextField(default=b'<csw:Record xmlns:csw="http://www.opengis.net/cat/2.0.2"/>', null=True, blank=True)),
                ('name', models.CharField(max_length=255, null=True, blank=True)),
                ('bbox_x0', models.DecimalField(null=True, max_digits=19, decimal_places=10, blank=True)),
                ('bbox_x1', models.DecimalField(null=True, max_digits=19, decimal_places=10, blank=True)),
                ('bbox_y0', models.DecimalField(null=True, max_digits=19, decimal_places=10, blank=True)),
                ('bbox_y1', models.DecimalField(null=True, max_digits=19, decimal_places=10, blank=True)),
                ('thumbnail', models.ImageField(null=True, upload_to=b'layers', blank=True)),
                ('page_url', models.URLField(max_length=255, null=True, blank=True)),
                ('is_monitored', models.BooleanField(default=True)),
                ('catalog', models.ForeignKey(default=1, editable=False, to='aggregator.Catalog')),
                ('keywords', taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='LayerDate',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date', models.CharField(max_length=25)),
                ('type', models.IntegerField(choices=[(0, b'Detected'), (1, b'From Metadata')])),
                ('layer', models.ForeignKey(to='aggregator.Layer')),
            ],
        ),
        migrations.CreateModel(
            name='LayerWM',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('category', models.CharField(max_length=255, null=True, blank=True)),
                ('username', models.CharField(max_length=255, null=True, blank=True)),
                ('temporal_extent_start', models.CharField(max_length=255, null=True, blank=True)),
                ('temporal_extent_end', models.CharField(max_length=255, null=True, blank=True)),
                ('layer', models.OneToOneField(to='aggregator.Layer')),
            ],
            options={
                'verbose_name': 'WorldMap Layer Attributes',
                'verbose_name_plural': 'WorldMap Layers Attributes',
            },
        ),
        migrations.CreateModel(
            name='Service',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(max_length=255, null=True, blank=True)),
                ('abstract', models.TextField(null=True, blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_updated', models.DateTimeField(auto_now=True)),
                ('active', models.BooleanField(default=True)),
                ('url', models.URLField(max_length=255)),
                ('is_public', models.BooleanField(default=True)),
                ('type', models.CharField(max_length=32, choices=[(b'OGC:CSW', b'Catalogue Service for the Web (CSW)'), (b'OGC:WMS', b'Web Map Service (WMS)'), (b'OGC:WMTS', b'Web Map Tile Service (WMTS)'), (b'OSGeo:TMS', b'Tile Map Service (TMS)'), (b'ESRI:ArcGIS:MapServer', b'ArcGIS REST MapServer'), (b'ESRI:ArcGIS:ImageServer', b'ArcGIS REST ImageServer'), (b'Hypermap:WorldMap', b'Harvard WorldMap'), (b'Hypermap:WARPER', b'Mapwarper')])),
                ('temporal_extent_start', models.CharField(max_length=255, null=True, blank=True)),
                ('temporal_extent_end', models.CharField(max_length=255, null=True, blank=True)),
                ('csw_last_updated', models.CharField(default=b'2016-08-15T20:00:11Z', max_length=32, null=True, blank=True)),
                ('csw_type', models.CharField(default=b'dataset', max_length=32)),
                ('csw_typename', models.CharField(default=b'csw:Record', max_length=32)),
                ('csw_schema', models.CharField(default=b'http://www.opengis.net/cat/csw/2.0.2', max_length=64)),
                ('anytext', models.TextField(null=True, blank=True)),
                ('wkt_geometry', models.TextField(default=b'POLYGON((-180 -90,-180 90,180 90,180 -90,-180 -90))')),
                ('xml', models.TextField(default=b'<csw:Record xmlns:csw="http://www.opengis.net/cat/2.0.2"/>', null=True, blank=True)),
                ('catalog', models.ForeignKey(default=1, editable=False, to='aggregator.Catalog')),
                ('keywords', taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SpatialReferenceSystem',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('code', models.CharField(max_length=255, null=True, blank=True)),
            ],
        ),
        migrations.CreateModel(
            name='TaskError',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('task_name', models.CharField(max_length=255)),
                ('args', models.CharField(max_length=255)),
                ('error_datetime', models.DateTimeField(auto_now=True)),
                ('message', models.TextField(null=True, blank=True)),
            ],
        ),
        migrations.AddField(
            model_name='service',
            name='srs',
            field=models.ManyToManyField(to='aggregator.SpatialReferenceSystem', blank=True),
        ),
        migrations.AddField(
            model_name='layer',
            name='service',
            field=models.ForeignKey(blank=True, to='aggregator.Service', null=True),
        ),
        migrations.AddField(
            model_name='endpoint',
            name='endpoint_list',
            field=models.ForeignKey(blank=True, to='aggregator.EndpointList', null=True),
        ),
        migrations.AlterUniqueTogether(
            name='endpoint',
            unique_together=set([('url', 'catalog')]),
        ),
    ]
